from __future__ import annotations
from dataclasses import dataclass
from typing import Union
from pathlib import Path
from functools import cached_property
from types import SimpleNamespace
from typing import TYPE_CHECKING
import numpy as np

from ase.units import Ha, Bohr

import gpaw.mpi as mpi
from gpaw.ibz2bz import IBZ2BZMaps
from gpaw.response.paw import LeanPAWDataset

if TYPE_CHECKING:
    from gpaw.new.ase_interface import ASECalculator
    from gpaw.calculator import GPAW
    from gpaw.setup import Setups, LeanSetup


class PAWDatasetCollection:
    def __init__(self, setups: Setups):
        by_species = {}
        by_atom = []
        id_by_atom = []

        for atom_id, setup in enumerate(setups):
            species_id = setups.id_a[atom_id]
            if species_id not in by_species:
                by_species[species_id] = ResponsePAWDataset(setup)
            by_atom.append(by_species[species_id])
            id_by_atom.append(species_id)

        self.by_species = by_species
        self.by_atom = by_atom
        self.id_by_atom = id_by_atom


GPWFilename = Union[Path, str]


class ResponseGroundStateAdapter:
    def __init__(self, calc: ASECalculator | GPAW):
        wfs = calc.wfs  # wavefunction object from gpaw.wavefunctions

        self.atoms = calc.atoms
        self.kd = wfs.kd  # KPointDescriptor object from gpaw.kpt_descriptor.
        self.world = calc.world  # _Communicator object from gpaw.mpi

        # GridDescriptor from gpaw.grid_descriptor.
        # Describes a grid in real space
        self.gd = wfs.gd

        # Also a GridDescriptor, with a finer grid...
        self.finegd = calc.density.finegd
        self.bd = wfs.bd  # BandDescriptor from gpaw.band_descriptor
        self.nspins = wfs.nspins  # number of spins: int
        self.dtype = wfs.dtype  # data type of wavefunctions, real or complex

        self.spos_ac = calc.spos_ac  # scaled position vector: np.ndarray

        self.kpt_u = wfs.kpt_u  # kpoints: list of Kpoint from gpaw.kpoint
        self.kpt_qs = wfs.kpt_qs  # kpoints: list of Kpoint from gpaw.kpoint

        self.fermi_level = wfs.fermi_level  # float
        self.atoms = calc.atoms  # ASE Atoms object
        self.pawdatasets = PAWDatasetCollection(calc.setups)

        self.pbc = self.atoms.pbc
        self.volume = self.gd.volume

        self.nvalence = wfs.nvalence
        self.nocc1, self.nocc2 = self.count_occupied_bands()

        self.ibz2bz = IBZ2BZMaps.from_calculator(calc)

        self._wfs = wfs
        self._density = calc.density
        self._hamiltonian = calc.hamiltonian
        self._calc = calc

    @classmethod
    def from_gpw_file(cls, gpw: GPWFilename) -> ResponseGroundStateAdapter:
        """Initiate the ground state adapter directly from a .gpw file."""
        from gpaw import GPAW, disable_dry_run
        assert Path(gpw).is_file()
        with disable_dry_run():
            calc = GPAW(gpw, txt=None, communicator=mpi.serial_comm)
        return cls(calc)

    @property
    def pd(self):
        # This is an attribute error in FD/LCAO mode.
        # We need to abstract away "calc" in all places used by response
        # code, and that includes places that are also compatible with FD.
        return self._wfs.pd

    @cached_property
    def global_pd(self):
        """Get a PWDescriptor that includes all k-points.

        In particular, this is necessary to allow all cores to be able to work
        on all k-points in the case where calc is parallelized over k-points,
        see gpaw.response.kspair
        """
        from gpaw.pw.descriptor import PWDescriptor

        assert self.gd.comm.size == 1
        kd = self.kd.copy()  # global KPointDescriptor without a comm
        return PWDescriptor(self.pd.ecut, self.gd,
                            dtype=self.pd.dtype,
                            kd=kd, fftwflags=self.pd.fftwflags,
                            gammacentered=self.pd.gammacentered)

    def get_occupations_width(self):
        # Ugly hack only used by pair.intraband_pair_density I think.
        # Actually: was copy-pasted in chi0 also.
        # More duplication can probably be eliminated around those.

        # Only works with Fermi-Dirac distribution
        occs = self._wfs.occupations
        assert occs.name in {'fermi-dirac', 'zero-width'}

        # No carriers when T=0
        width = getattr(occs, '_width', 0.0) / Ha
        return width

    @cached_property
    def cd(self):
        return CellDescriptor(self.gd.cell_cv, self.pbc)

    @property
    def nt_sR(self):
        # Used by localft and fxc_kernels
        return self._density.nt_sG

    @property
    def nt_sr(self):
        # Used by localft
        if self._density.nt_sg is None:
            self._density.interpolate_pseudo_density()
        return self._density.nt_sg

    @cached_property
    def n_sR(self):
        return self._density.get_all_electron_density(
            atoms=self.atoms, gridrefinement=1)[0]

    @cached_property
    def n_sr(self):
        return self._density.get_all_electron_density(
            atoms=self.atoms, gridrefinement=2)[0]

    @property
    def D_asp(self):
        # Used by fxc_kernels
        return self._density.D_asp

    def get_pseudo_density(self, gridrefinement=2):
        # Used by localft
        if gridrefinement == 1:
            return self.nt_sR, self.gd
        elif gridrefinement == 2:
            return self.nt_sr, self.finegd
        else:
            raise ValueError(f'Invalid gridrefinement {gridrefinement}')

    def get_all_electron_density(self, gridrefinement=2):
        # Used by fxc, fxc_kernels and localft
        if gridrefinement == 1:
            return self.n_sR, self.gd
        elif gridrefinement == 2:
            return self.n_sr, self.finegd
        else:
            raise ValueError(f'Invalid gridrefinement {gridrefinement}')

    # Things used by EXX.  This is getting pretty involved.
    #
    # EXX naughtily accesses the density object in order to
    # interpolate_pseudo_density() which is in principle mutable.

    def hacky_all_electron_density(self, **kwargs):
        # fxc likes to get all electron densities.  It calls
        # calc.get_all_electron_density() and so we wrap that here.
        # But it also collects to serial (bad), and it also zeropads
        # nonperiodic directions (probably WRONG!).
        #
        # Also this one returns in user units, whereas the calling
        # code actually wants internal units.  Very silly then.
        #
        # ALso, the calling code often wants the gd, which is not
        # returned, so it is redundantly reconstructed in multiple
        # places by refining the "right" number of times.
        n_g = self._calc.get_all_electron_density(**kwargs)
        n_g *= Bohr**3
        return n_g

    # Used by EXX.
    @property
    def hamiltonian(self):
        return self._hamiltonian

    # Used by EXX.
    @property
    def density(self):
        return self._density

    # Ugh SOC
    def soc_eigenstates(self, **kwargs):
        from gpaw.spinorbit import soc_eigenstates
        return soc_eigenstates(self._calc, **kwargs)

    @property
    def xcname(self):
        return self.hamiltonian.xc.name

    def get_xc_difference(self, xc):
        # XXX used by gpaw/xc/tools.py
        return self._calc.get_xc_difference(xc)

    def get_wave_function_array(self, u, n):
        # XXX used by gpaw/xc/tools.py in a hacky way
        return self._wfs._get_wave_function_array(
            u, n, realspace=True)

    def pair_density_paw_corrections(self, qpd):
        from gpaw.response.paw import get_pair_density_paw_corrections
        return get_pair_density_paw_corrections(
            pawdatasets=self.pawdatasets, qpd=qpd, spos_ac=self.spos_ac,
            atomrotations=self.atomrotations)

    def matrix_element_paw_corrections(self, qpd, rshe_a):
        from gpaw.response.paw import get_matrix_element_paw_corrections
        return get_matrix_element_paw_corrections(
            qpd, self.pawdatasets, rshe_a, self.spos_ac)

    def get_pos_av(self):
        # gd.cell_cv must always be the same as pd.gd.cell_cv, right??
        return np.dot(self.spos_ac, self.gd.cell_cv)

    def count_occupied_bands(self, ftol=1e-6):
        """Count the number of filled and non-empty bands.

        ftol : float
            Threshold determining whether a band is completely filled
            (f > 1 - ftol) or completely empty (f < ftol).
        """
        nocc1 = 9999999
        nocc2 = 0
        for kpt in self.kpt_u:
            f_n = kpt.f_n / kpt.weight
            nocc1 = min((f_n > 1 - ftol).sum(), nocc1)
            nocc2 = max((f_n > ftol).sum(), nocc2)
        return nocc1, nocc2

    def get_eigenvalue_range(self, nbands: int | None = None):
        """Get smallest and largest Kohn-Sham eigenvalues."""
        if nbands is None:
            nbands = self.bd.nbands
        assert 1 <= nbands <= self.bd.nbands

        epsmin = np.inf
        epsmax = -np.inf
        for kpt in self.kpt_u:
            epsmin = min(epsmin, kpt.eps_n[0])  # the eigenvalues are ordered
            epsmax = max(epsmax, kpt.eps_n[nbands - 1])
        return epsmin, epsmax

    @property
    def metallic(self):
        # Does the number of filled bands equal the number of non-empty bands?
        return self.nocc1 != self.nocc2

    @cached_property
    def ibzq_qc(self):
        # For G0W0Kernel
        kd = self.kd
        bzq_qc = kd.get_bz_q_points(first=True)
        U_scc = kd.symmetry.op_scc
        ibzq_qc = kd.get_ibz_q_points(bzq_qc, U_scc)[0]

        return ibzq_qc

    def get_ibz_vertices(self):
        # For the tetrahedron method in Chi0
        from gpaw.bztools import get_bz
        # NB: We are ignoring the pbc_c keyword to get_bz() in order to mimic
        # find_high_symmetry_monkhorst_pack() in gpaw.bztools. XXX
        _, ibz_vertices_kc, _ = get_bz(self._calc)
        return ibz_vertices_kc

    def get_aug_radii(self):
        return np.array([max(pawdata.rcut_j)
                         for pawdata in self.pawdatasets.by_atom])

    @cached_property
    def micro_setups(self):
        from gpaw.response.localft import extract_micro_setup
        micro_setups = []
        for a, pawdata in enumerate(self.pawdatasets.by_atom):
            micro_setups.append(extract_micro_setup(pawdata, self.D_asp[a]))
        return micro_setups

    @property
    def atomrotations(self):
        return self._wfs.setups.atomrotations

    @cached_property
    def kpoints(self):
        from gpaw.response.kpoints import ResponseKPointGrid
        return ResponseKPointGrid(self.kd, self.gd.icell_cv, self.kd.bzk_kc)


@dataclass
class CellDescriptor:
    cell_cv: np.ndarray
    pbc_c: np.ndarray

    @property
    def nonperiodic_hypervolume(self):
        """Get the hypervolume of the cell along nonperiodic directions.

        Returns the hypervolume Λ in units of Å, where

        Λ = 1        in 3D
        Λ = L        in 2D, where L is the out-of-plane cell vector length
        Λ = A        in 1D, where A is the transverse cell area
        Λ = V        in 0D, where V is the cell volume
        """
        cell_cv = self.cell_cv
        pbc_c = self.pbc_c
        if sum(pbc_c) > 0:
            # In 1D and 2D, we assume the cartesian representation of the unit
            # cell to be block diagonal, separating the periodic and
            # nonperiodic cell vectors in different blocks.
            assert np.allclose(cell_cv[~pbc_c][:, pbc_c], 0.) and \
                np.allclose(cell_cv[pbc_c][:, ~pbc_c], 0.), \
                "In 1D and 2D, please put the periodic/nonperiodic axis " \
                "along a cartesian component"
        L = np.abs(np.linalg.det(cell_cv[~pbc_c][:, ~pbc_c]))
        return L * Bohr**sum(~pbc_c)  # Bohr -> Å


# Contains all the relevant information
# from Setups class for response calculators
class ResponsePAWDataset(LeanPAWDataset):
    def __init__(self, setup: LeanSetup, **kwargs):
        super().__init__(
            rgd=setup.rgd, l_j=setup.l_j, rcut_j=setup.rcut_j,
            phit_jg=setup.data.phit_jg, phi_jg=setup.data.phi_jg, **kwargs)
        assert setup.ni == self.ni

        self.n_j = setup.n_j
        self.N0_q = setup.N0_q
        self.nabla_iiv = setup.nabla_iiv
        self.xc_correction: SimpleNamespace | None
        if setup.xc_correction is not None:
            self.xc_correction = SimpleNamespace(
                rgd=setup.xc_correction.rgd, Y_nL=setup.xc_correction.Y_nL,
                n_qg=setup.xc_correction.n_qg, nt_qg=setup.xc_correction.nt_qg,
                nc_g=setup.xc_correction.nc_g, nct_g=setup.xc_correction.nct_g,
                nc_corehole_g=setup.xc_correction.nc_corehole_g,
                B_pqL=setup.xc_correction.B_pqL,
                e_xc0=setup.xc_correction.e_xc0)
        else:
            # If there is no `xc_correction` in the setup, we assume to be
            # using pseudo potentials.
            self.xc_correction = None
            # In this case, we set l_j to an empty list in order to bypass the
            # calculation of PAW corrections to pair densities etc.
            # This is quite an ugly hack...
            # If we want to support pseudo potential calculations for real, we
            # should skip the PAW corrections at the matrix element calculator
            # level, not by an odd hack.
            self.l_j = np.array([], dtype=float)
        self.hubbard_u = setup.hubbard_u



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
import sys

import numpy as np
from ase.units import Hartree

import gpaw.mpi as mpi

from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.dyson import DysonEquation
from gpaw.response.density_kernels import get_density_xc_kernel
from gpaw.response.chi0 import Chi0Calculator, get_frequency_descriptor
from gpaw.response.chi0_data import Chi0Data
from gpaw.response.pair import get_gs_and_context

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gpaw.response.groundstate import CellDescriptor
    from gpaw.response.frequencies import FrequencyDescriptor
    from gpaw.response.pair_functions import SingleQPWDescriptor


"""
On the notation in this module.

When calculating properties such as the dielectric function, EELS spectrum and
polarizability there are many inherent subtleties relating to (ir)reducible
representations and inclusion of local-field effects. For the reciprocal space
representation of the Coulomb potential, we use the following notation

v or v(q): The bare Coulomb interaction, 4π/|G+q|²

V or V(q): The specified Coulomb interaction. Will usually be either the bare
           interaction or a truncated version hereof.
ˍ    ˍ
V or V(q): The modified Coulomb interaction. Equal to V(q) for finite
           reciprocal wave vectors G > 0, but modified to exclude long-range
           interactions, that is, equal to 0 for G = 0.
"""


@dataclass
class Chi0DysonEquations:
    chi0: Chi0Data
    df: 'DielectricFunctionCalculator'

    def __post_init__(self):
        self.gs = self.df.gs
        self.context = self.df.context
        self.coulomb = self.df.coulomb
        if self.coulomb.truncation is None:
            self.bare_coulomb = self.coulomb
        else:
            self.bare_coulomb = self.coulomb.new(truncation=None)
        # When inverting the Dyson equation, we distribute frequencies globally
        blockdist = self.chi0.body.blockdist.new_distributor(nblocks='max')
        self.wblocks = Blocks1D(blockdist.blockcomm, len(self.chi0.wd))

    @staticmethod
    def _normalize(direction):
        if isinstance(direction, str):
            d_v = {'x': [1, 0, 0],
                   'y': [0, 1, 0],
                   'z': [0, 0, 1]}[direction]
        else:
            d_v = direction
        d_v = np.asarray(d_v) / np.linalg.norm(d_v)
        return d_v

    def get_chi0_wGG(self, direction='x'):
        chi0 = self.chi0
        chi0_wGG = chi0.body.get_distributed_frequencies_array().copy()
        if chi0.qpd.optical_limit:
            # Project head and wings along the input direction
            d_v = self._normalize(direction)
            W_w = self.wblocks.myslice
            chi0_wGG[:, 0] = np.dot(d_v, chi0.chi0_WxvG[W_w, 0])
            chi0_wGG[:, :, 0] = np.dot(d_v, chi0.chi0_WxvG[W_w, 1])
            chi0_wGG[:, 0, 0] = np.dot(d_v, np.dot(chi0.chi0_Wvv[W_w], d_v).T)
        return chi0_wGG

    def get_Kxc_GG(self, *, xc, chi0_wGG, **kwargs):
        """Get adiabatic xc kernel (TDDFT).

        Choose between ALDA, Bootstrap and LRalpha (long-range kernel), where
        alpha is a user specified parameter (for example xc='LR0.25')."""
        return get_density_xc_kernel(
            self.chi0.qpd, self.gs, self.context,
            functional=xc, chi0_wGG=chi0_wGG, **kwargs)

    def get_coulomb_scaled_kernel(self, xc='RPA', modified=False, **xckwargs):
        """Get the Hxc kernel rescaled by the bare Coulomb potential v(q).

        Calculates
        ˷
        K(q) = v^(-1/2)(q) K_Hxc(q) v^(-1/2)(q),

        where v(q) is the bare Coulomb potential and

        K_Hxc(q) = V(q) + K_xc(q).

        When using the `modified` flag, the specified Coulomb kernel will be
        replaced with its modified analogue,
                ˍ
        V(q) -> V(q)
        """
        qpd = self.chi0.qpd
        if self.coulomb is self.bare_coulomb:
            v_G = self.coulomb.V(qpd)  # bare Coulomb interaction
            K_GG = np.eye(len(v_G), dtype=complex)
        else:
            v_G = self.bare_coulomb.V(qpd)
            V_G = self.coulomb.V(qpd)
            K_GG = np.diag(V_G / v_G)
        if modified:
            K_GG[0, 0] = 0.
        if xc != 'RPA':
            Kxc_GG = self.get_Kxc_GG(xc=xc, **xckwargs)
            sqrtv_G = v_G**0.5
            K_GG += Kxc_GG / sqrtv_G / sqrtv_G[:, np.newaxis]
        return v_G, K_GG

    @staticmethod
    def invert_dyson_like_equation(in_wGG, K_GG, reuse_buffer=True):
        """Generalized Dyson equation invertion.

        Calculates

        B(q,ω) = [1 - A(q,ω) K(q)]⁻¹ A(q,ω)

        while possibly storing the output B(q,ω) in the input A(q,ω) buffer.
        """
        if reuse_buffer:
            out_wGG = in_wGG
        else:
            out_wGG = np.zeros_like(in_wGG)
        for w, in_GG in enumerate(in_wGG):
            out_wGG[w] = DysonEquation(in_GG, in_GG @ K_GG).invert()
        return out_wGG

    def rpa_density_response(self, direction='x', qinf_v=None):
        """Calculate the RPA susceptibility for (semi-)finite q."""
        # Extract χ₀(q,ω)
        qpd = self.chi0.qpd
        chi0_wGG = self.get_chi0_wGG(direction=direction)
        if qpd.optical_limit:
            # Restore the q-dependence of the head and wings in the q→0 limit
            assert qinf_v is not None and np.linalg.norm(qinf_v) > 0.
            d_v = self._normalize(direction)
            chi0_wGG[:, 1:, 0] *= np.dot(qinf_v, d_v)
            chi0_wGG[:, 0, 1:] *= np.dot(qinf_v, d_v)
            chi0_wGG[:, 0, 0] *= np.dot(qinf_v, d_v)**2
        # Invert Dyson equation, χ(q,ω) = [1 - χ₀(q,ω) V(q)]⁻¹ χ₀(q,ω)
        V_GG = self.coulomb.kernel(qpd, q_v=qinf_v)
        chi_wGG = self.invert_dyson_like_equation(chi0_wGG, V_GG)
        return qpd, chi_wGG, self.wblocks

    def inverse_dielectric_function(self, *args, **kwargs):
        """Calculate v^(1/2) χ v^(1/2), from which ε⁻¹(q,ω) is constructed."""
        return InverseDielectricFunction.from_chi0_dyson_eqs(
            self, *self.calculate_vchi_symm(*args, **kwargs))

    def calculate_vchi_symm(self, xc='RPA', direction='x', **kwargs):
        """Calculate v^(1/2) χ v^(1/2).

        Starting from the TDDFT Dyson equation

        χ(q,ω) = χ₀(q,ω) + χ₀(q,ω) K_Hxc(q,ω) χ(q,ω),                (1)

        the Coulomb scaled susceptibility,
        ˷
        χ(q,ω) = v^(1/2)(q) χ(q,ω) v^(1/2)(q)

        can be calculated from the Dyson-like equation
        ˷        ˷         ˷       ˷      ˷
        χ(q,ω) = χ₀(q,ω) + χ₀(q,ω) K(q,ω) χ(q,ω)                     (2)

        where
        ˷
        K(q,ω) = v^(-1/2)(q) K_Hxc(q,ω) v^(-1/2)(q).

        Here v(q) refers to the bare Coulomb potential. It should be emphasized
        that invertion of (2) rather than (1) is not merely a rescaling
        excercise. In the optical q → 0 limit, the Coulomb kernel v(q) diverges
        as 1/|G+q|² while the Kohn-Sham susceptibility χ₀(q,ω) vanishes as
        |G+q|². Treating v^(1/2)(q) χ₀(q,ω) v^(1/2)(q) as a single variable,
        the effects of this cancellation can be treated accurately within k.p
        perturbation theory.
        """
        chi0_wGG = self.get_chi0_wGG(direction=direction)
        v_G, K_GG = self.get_coulomb_scaled_kernel(
            xc=xc, chi0_wGG=chi0_wGG, **kwargs)
        # Calculate v^(1/2)(q) χ₀(q,ω) v^(1/2)(q)
        sqrtv_G = v_G**0.5
        vchi0_symm_wGG = chi0_wGG  # reuse buffer
        for w, chi0_GG in enumerate(chi0_wGG):
            vchi0_symm_wGG[w] = chi0_GG * sqrtv_G * sqrtv_G[:, np.newaxis]
        # Invert Dyson equation
        vchi_symm_wGG = self.invert_dyson_like_equation(
            vchi0_symm_wGG, K_GG, reuse_buffer=False)
        return vchi0_symm_wGG, vchi_symm_wGG

    def customized_dielectric_function(self, *args, **kwargs):
        """Calculate Ε(q,ω) = 1 - V(q) P(q,ω)."""
        V_GG = self.coulomb.kernel(self.chi0.qpd)
        P_wGG = self.polarizability_operator(*args, **kwargs)
        nG = len(V_GG)
        eps_wGG = P_wGG  # reuse buffer
        for w, P_GG in enumerate(P_wGG):
            eps_wGG[w] = np.eye(nG) - V_GG @ P_GG
        return CustomizableDielectricFunction.from_chi0_dyson_eqs(
            self, eps_wGG)

    def bare_dielectric_function(self, xc='RPA', direction='x', **ignored):
        """Calculate v^(1/2) ̄χ v^(1/2), from which ̄ϵ=1-v ̄χ is constructed.

        The unscreened susceptibility is given by the Dyson-like equation
        ˍ                        ˍ    ˍ
        χ(q,ω) = P(q,ω) + P(q,ω) V(q) χ(q,ω),                        (3)

        In the special case of RPA, where P(q,ω) = χ₀(q,ω), one may notice that
        the Dyson-like equation (3) is exactly identical to the TDDFT Dyson
        equation (1) when replacing the Hartree-exchange-correlation kernel
        with the modified Coulomb interaction:
                    ˍ
        K_Hxc(q) -> V(q).
                                                                  ˍ
        We may thus reuse that functionality to calculate v^(1/2) χ v^(1/2).
        """
        if xc != 'RPA':
            raise NotImplementedError(
                'Calculation of the bare dielectric function within TDDFT has '
                'not yet been implemented. For TDDFT dielectric properties, '
                'please calculate the inverse dielectric function.')
        vP_symm_wGG, vchibar_symm_wGG = self.calculate_vchi_symm(
            xc=xc, direction=direction, modified=True)
        return BareDielectricFunction.from_chi0_dyson_eqs(
            self, vP_symm_wGG, vchibar_symm_wGG)

    def polarizability_operator(self, xc='RPA', direction='x', **xckwargs):
        """Calculate the polarizability operator P(q,ω).

        Depending on the theory (RPA, TDDFT, MBPT etc.), the polarizability
        operator is approximated in various ways see e.g.
        [Rev. Mod. Phys. 74, 601 (2002)].

        In RPA:
            P(q,ω) = χ₀(q,ω)

        In TDDFT:
            P(q,ω) = [1 - χ₀(q,ω) K_xc(q,ω)]⁻¹ χ₀(q,ω)
        """
        chi0_wGG = self.get_chi0_wGG(direction=direction)
        if xc == 'RPA':
            return chi0_wGG
        # TDDFT (in adiabatic approximations to the kernel)
        if self.chi0.qpd.optical_limit:
            raise NotImplementedError(
                'Calculation of the TDDFT dielectric function via the '
                'polarizability operator has not been implemented for the '
                'optical limit. Please calculate the inverse dielectric '
                'function instead.')
            # The TDDFT implementation here is invalid in the optical limit
            # since the chi0_wGG variable already contains Coulomb effects,
            # chi0_wGG ~ V^(1/2) χ₀ V^(1/2)/(4π).
            # Furthermore, a direct evaluation of V(q) P(q,ω) does not seem
            # sensible, since it does not account for the exact cancellation
            # of the q-dependences of the two functions.
            # In principle, one could treat the v(q) P(q,ω) product in
            # perturbation theory, similar to the χ₀(q,ω) v(q) product in the
            # Dyson equation for χ, but unless we need to calculate the TDDFT
            # polarizability using truncated kernels, this isn't really
            # necessary.
        Kxc_GG = self.get_Kxc_GG(xc=xc, chi0_wGG=chi0_wGG, **xckwargs)
        return self.invert_dyson_like_equation(chi0_wGG, Kxc_GG)


@dataclass
class DielectricFunctionData(ABC):
    cd: CellDescriptor
    qpd: SingleQPWDescriptor
    wd: FrequencyDescriptor
    wblocks: Blocks1D
    coulomb: CoulombKernel
    bare_coulomb: CoulombKernel

    @classmethod
    def from_chi0_dyson_eqs(cls, chi0_dyson_eqs, *args, **kwargs):
        chi0 = chi0_dyson_eqs.chi0
        return cls(chi0_dyson_eqs.gs.cd, chi0.qpd,
                   chi0.wd, chi0_dyson_eqs.wblocks,
                   chi0_dyson_eqs.coulomb, chi0_dyson_eqs.bare_coulomb,
                   *args, **kwargs)

    def _macroscopic_component(self, in_wGG):
        return self.wblocks.all_gather(in_wGG[:, 0, 0])

    @property
    def v_G(self):
        return self.bare_coulomb.V(self.qpd)

    @abstractmethod
    def macroscopic_dielectric_function(self) -> ScalarResponseFunctionSet:
        """Get the macroscopic dielectric function ε_M(q,ω)."""

    def dielectric_constant(self):
        return self.macroscopic_dielectric_function().static_limit.real

    def eels_spectrum(self):
        """Get the macroscopic EELS spectrum.

        The spectrum is defined as

                                          1
        EELS(q,ω) ≡ -Im ε⁻¹(q,ω) = -Im ‾‾‾‾‾‾‾.
                         00            ε (q,ω)
                                        M

        In addition to the many-body spectrum, we also calculate the
        EELS spectrum in the relevant independent-particle approximation,
        here defined as

                          1
        EELS₀(ω) = -Im ‾‾‾‾‾‾‾.
                        IP
                       ε (q,ω)
                        M
        """
        _, eps0_W, eps_W = self.macroscopic_dielectric_function().arrays
        eels0_W = -(1. / eps0_W).imag
        eels_W = -(1. / eps_W).imag
        return ScalarResponseFunctionSet(self.wd, eels0_W, eels_W)

    def polarizability(self):
        """Get the macroscopic polarizability α_M(q,ω).

        In the special case where dielectric properties are calculated
        solely based on the bare Coulomb interaction V(q) = v(q), the
        macroscopic part of the electronic polarizability α(q,ω) is related
        directly to the macroscopic dielectric function ε_M(q,ω).

        In particular, we define the macroscroscopic polarizability so as to
        make it independent of the nonperiodic cell vector lengths. Namely,

        α (q,ω) ≡ Λ α (q,ω)
         M           00

        where Λ is the nonperiodic hypervolume of the unit cell. Thus,

        α_M(q,ω) = Λ/(4π) (ϵ_M(q,ω) - 1) = Λ/(4π) (ε_M(q,ω) - 1),

        where the latter equality only holds in the special case V(q) = v(q).

        In addition to α_M(q,ω), we calculate also the macroscopic
        polarizability in the relevant independent-particle approximation by
        replacing ϵ_M with ε_M^(IP).
        """
        if self.coulomb is not self.bare_coulomb:
            raise ValueError(
                'When using a truncated Coulomb kernel, the polarizability '
                'cannot be calculated based on the macroscopic dielectric '
                'function. Please calculate the bare dielectric function '
                'instead.')
        _, eps0_W, eps_W = self.macroscopic_dielectric_function().arrays
        return self._polarizability(eps0_W, eps_W)

    def _polarizability(self, eps0_W, eps_W):
        L = self.cd.nonperiodic_hypervolume
        alpha0_W = L / (4 * np.pi) * (eps0_W - 1)
        alpha_W = L / (4 * np.pi) * (eps_W - 1)
        return ScalarResponseFunctionSet(self.wd, alpha0_W, alpha_W)


@dataclass
class InverseDielectricFunction(DielectricFunctionData):
    """Data class for the inverse dielectric function ε⁻¹(q,ω).

    The inverse dielectric function characterizes the longitudinal response

    V (q,ω) = ε⁻¹(q,ω) V (q,ω),
     tot                ext

    where the induced potential due to the electronic system is given by vχ,

    ε⁻¹(q,ω) = 1 + v(q) χ(q,ω).

    In this data class, ε⁻¹ is cast in terms if its symmetrized representation
    ˷
    ε⁻¹(q,ω) = v^(-1/2)(q) ε⁻¹(q,ω) v^(1/2)(q),

    that is, in terms of v^(1/2)(q) χ(q,ω) v^(1/2)(q).

    Please remark that v(q) here refers to the bare Coulomb potential
    irregardless of whether χ(q,ω) was determined using a truncated analogue.
    """
    vchi0_symm_wGG: np.ndarray  # v^(1/2)(q) χ₀(q,ω) v^(1/2)(q)
    vchi_symm_wGG: np.ndarray

    def macroscopic_components(self):
        vchi0_W = self._macroscopic_component(self.vchi0_symm_wGG)
        vchi_W = self._macroscopic_component(self.vchi_symm_wGG)
        return vchi0_W, vchi_W

    def macroscopic_dielectric_function(self):
        """Get the macroscopic dielectric function ε_M(q,ω).

        Calculates

                       1
        ε (q,ω) =  ‾‾‾‾‾‾‾‾
         M         ε⁻¹(q,ω)
                    00

        along with the macroscopic dielectric function in the independent-
        particle random-phase approximation [Rev. Mod. Phys. 74, 601 (2002)],

         IPRPA
        ε (q,ω) = 1 - v(q) χ⁰(q,ω)
         M                  00

        that is, neglecting local-field and exchange-correlation effects.
        """
        vchi0_W, vchi_W = self.macroscopic_components()
        eps0_W = 1 - vchi0_W
        eps_W = 1 / (1 + vchi_W)
        return ScalarResponseFunctionSet(self.wd, eps0_W, eps_W)

    def dynamic_susceptibility(self):
        """Get the macroscopic components of χ(q,ω) and χ₀(q,ω)."""
        vchi0_W, vchi_W = self.macroscopic_components()
        v0 = self.v_G[0]  # Macroscopic Coulomb potential (4π/q²)
        return ScalarResponseFunctionSet(self.wd, vchi0_W / v0, vchi_W / v0)


@dataclass
class CustomizableDielectricFunction(DielectricFunctionData):
    """Data class for customized dielectric functions Ε(q,ω).

    Ε(q,ω) is customizable in the sense that bare Coulomb interaction v(q) is
    replaced by an arbitrary interaction V(q) in the formula for ε(q,ω),

    Ε(q,ω) = 1 - V(q) P(q,ω),

    where P is the polarizability operator [Rev. Mod. Phys. 74, 601 (2002)].
    Thus, for any truncated or otherwise cusomized interaction V(q) ≠ v(q),
    Ε(q,ω) ≠ ε(q,ω) and Ε⁻¹(q,ω) ≠ ε⁻¹(q,ω).
    """
    eps_wGG: np.ndarray

    def macroscopic_customized_dielectric_function(self):
        """Get the macroscopic customized dielectric function Ε_M(q,ω).

        We define the macroscopic customized dielectric function as

                      1
        Ε (q,ω) =  ‾‾‾‾‾‾‾‾,
         M         Ε⁻¹(q,ω)
                    00

        and calculate also the macroscopic dielectric function in the
        customizable independent-particle approximation:

         cIP
        ε (q,ω) = 1 - VP (q,ω) = Ε (q,ω).
         M              00        00
        """
        eps0_W = self._macroscopic_component(self.eps_wGG)
        # Invert Ε(q,ω) one frequency at a time to compute Ε_M(q,ω)
        eps_w = np.zeros((self.wblocks.nlocal,), complex)
        for w, eps_GG in enumerate(self.eps_wGG):
            eps_w[w] = 1 / np.linalg.inv(eps_GG)[0, 0]
        eps_W = self.wblocks.all_gather(eps_w)
        return ScalarResponseFunctionSet(self.wd, eps0_W, eps_W)

    def macroscopic_dielectric_function(self):
        """Get the macroscopic dielectric function ε_M(q,ω).

        In the special case V(q) = v(q), Ε_M(q,ω) = ε_M(q,ω).
        """
        if self.coulomb is not self.bare_coulomb:
            raise ValueError(
                'The macroscopic dielectric function is defined in terms of '
                'the bare Coulomb interaction. To truncate the Hartree '
                'electron-electron correlations, please calculate the inverse '
                'dielectric function instead.')
        return self.macroscopic_customized_dielectric_function()


@dataclass
class BareDielectricFunction(DielectricFunctionData):
    """Data class for the bare (unscreened) dielectric function.

    The bare dielectric function is defined in terms of the unscreened
    susceptibility,
    ˍ                 ˍ
    ϵ(q,ω) = 1 - v(q) χ(q,ω),
                                            ˍ
    here represented in terms of v^(1/2)(q) χ(q,ω) v^(1/2)(q).
    """
    vP_symm_wGG: np.ndarray  # v^(1/2) P v^(1/2)
    vchibar_symm_wGG: np.ndarray

    def macroscopic_components(self):
        vP_W = self._macroscopic_component(self.vP_symm_wGG)
        vchibar_W = self._macroscopic_component(self.vchibar_symm_wGG)
        return vP_W, vchibar_W

    def macroscopic_bare_dielectric_function(self):
        """Get the macroscopic bare dielectric function ϵ_M(q,ω).

        Calculates
                   ˍ
        ϵ_M(q,ω) = ϵ (q,ω)
                    00

        along with the macroscopic dielectric function in the independent-
        particle approximation:

         IP
        ε (q,ω) = 1 - vP (q,ω).
         M              00
        """
        vP_W, vchibar_W = self.macroscopic_components()
        eps0_W = 1. - vP_W
        eps_W = 1. - vchibar_W
        return ScalarResponseFunctionSet(self.wd, eps0_W, eps_W)

    def macroscopic_dielectric_function(self):
        """Get the macroscopic dielectric function ε_M(q,ω).

        In the special case where the unscreened susceptibility is calculated
        using the bare Coulomb interaction,
        ˍ                        ˍ    ˍ
        χ(q,ω) = P(q,ω) + P(q,ω) v(q) χ(q,ω),

        it holds identically that [Rev. Mod. Phys. 74, 601 (2002)]:

        ε_M(q,ω) = ϵ_M(q,ω).
        """
        if self.coulomb is not self.bare_coulomb:
            raise ValueError(
                'The macroscopic dielectric function cannot be obtained from '
                'the bare dielectric function calculated based on a truncated '
                'Coulomb interaction. Please calculate the inverse dielectric '
                'function instead')
        return self.macroscopic_bare_dielectric_function()

    def polarizability(self):
        """Get the macroscopic polarizability α_M(q,ω).

        In the most general case, the electronic polarizability can be defined
        in terms of the unscreened susceptibility
                               ˍ                ˍ
        α(q,ω) ≡ - 1/(4π) v(q) χ(q,ω) = 1/(4π) (ϵ(q,ω) - 1).

        This is especially useful in systems of reduced dimensionality, where
        χ(q,ω) needs to be calculated using a truncated Coulomb kernel in order
        to achieve convergence in a feasible way.
        """
        _, eps0_W, eps_W = self.macroscopic_bare_dielectric_function().arrays
        return self._polarizability(eps0_W, eps_W)


class DielectricFunctionCalculator:
    def __init__(self, chi0calc: Chi0Calculator, coulomb: CoulombKernel):
        self.chi0calc = chi0calc
        self.coulomb = coulomb

        self.gs = chi0calc.gs
        self.context = chi0calc.context

        self._chi0cache: dict = {}

    def calculate_chi0(self, q_c: list | np.ndarray) -> Chi0DysonEquations:
        """Calculates the response function.

        Calculate the response function for a specific momentum.

        q_c: [float, float, float]
            The momentum wavevector.
        """

        # We cache the computed data since chi0 may otherwise be redundantly
        # calculated e.g. if the user calculates multiple directions.
        #
        # May be called multiple times with same q_c, and we want to
        # be able to recognize previous seen values of q_c.
        # We do this by rounding and converting to string with fixed
        # precision (so not very elegant).
        q_key = [f'{q:.10f}' for q in q_c]
        key = tuple(q_key)

        if key not in self._chi0cache:
            # We assume that the caller will trigger this multiple
            # times with the same qpoint, then several times with
            # another qpoint, etc.  If that's true, then we
            # need to cache no more than one qpoint at a time.
            # Thus to save memory, we clear the cache here.
            #
            # This should be replaced with something more reliable,
            # such as having the caller manage things more explicitly.
            #
            # See https://gitlab.com/gpaw/gpaw/-/issues/662
            #
            # In conclusion, delete the cache now:
            self._chi0cache.clear()

            # cache Chi0Data from gpaw.response.chi0_data
            self._chi0cache[key] = Chi0DysonEquations(
                self.chi0calc.calculate(q_c), self)
            self.context.write_timer()

        return self._chi0cache[key]

    def _dielectric_function_new(self, q_c=[0, 0, 0], direction='x',
                                 **xckwargs) -> DielectricFunctionData:
        # Temporary method until truncation becomes a method input XXX
        chi0_dyson_eqs = self.calculate_chi0(q_c)
        if self.coulomb.truncation:
            # eps: BareDielectricFunction
            method = chi0_dyson_eqs.bare_dielectric_function
        else:
            # eps: CustomizableDielectricFunction
            method = chi0_dyson_eqs.customized_dielectric_function
        eps = method(direction=direction, **xckwargs)
        return eps

    def get_bare_dielectric_function(self, q_c=[0, 0, 0], direction='x',
                                     **xckwargs) -> BareDielectricFunction:
        return self.calculate_chi0(q_c).bare_dielectric_function(
            direction=direction, **xckwargs)

    def get_literal_dielectric_function(
            self, *args, **kwargs) -> CustomizableDielectricFunction:
        """Calculate the dielectric function ε(q,ω) = 1 - v(q) P(q,ω)."""
        return self.get_customized_dielectric_function(
            truncation=None, *args, **kwargs)

    def get_customized_dielectric_function(
            self, *, truncation: str | None,
            q_c=[0, 0, 0], direction='x',
            **xckwargs) -> CustomizableDielectricFunction:
        # NB: ignores self.coulomb while this still exists XXX
        chi0_dyson_equation = self.calculate_chi0(q_c)
        if truncation is None:
            chi0_dyson_equation.coulomb = chi0_dyson_equation.bare_coulomb
        else:
            chi0_dyson_equation.coulomb = chi0_dyson_equation.coulomb.new(
                truncation=truncation)
        return chi0_dyson_equation.customized_dielectric_function(
            direction=direction, **xckwargs)

    def get_inverse_dielectric_function(
            self, q_c=[0, 0, 0], direction='x',
            **xckwargs) -> InverseDielectricFunction:
        return self.calculate_chi0(q_c).inverse_dielectric_function(
            direction=direction, **xckwargs)

    def get_rpa_density_response(self, q_c, *, direction, qinf_v=None):
        # Used by the QEH code
        return self.calculate_chi0(q_c).rpa_density_response(
            direction=direction, qinf_v=qinf_v)


class DielectricFunction(DielectricFunctionCalculator):
    """This class defines dielectric function related physical quantities."""

    def __init__(self, calc, *,
                 frequencies=None,
                 ecut=50,
                 hilbert=True,
                 nbands=None, eta=0.2,
                 intraband=True, nblocks=1, world=mpi.world, txt=sys.stdout,
                 truncation=None, disable_point_group=False,
                 disable_time_reversal=False,
                 integrationmode=None, rate=0.0,
                 eshift: float | None = None):
        """Creates a DielectricFunction object.

        calc: str
            The ground-state calculation file that the linear response
            calculation is based on.
        frequencies:
            Input parameters for frequency_grid.
            Can be an array of frequencies to evaluate the response function at
            or dictionary of parameters for build-in nonlinear grid
            (see :ref:`frequency grid`).
        ecut: float
            Plane-wave cut-off.
        hilbert: bool
            Use hilbert transform.
        nbands: int
            Number of bands from calculation.
        eta: float
            Broadening parameter.
        intraband: bool
            Include intraband transitions.
        world: comm
            mpi communicator.
        nblocks: int
            Split matrices in nblocks blocks and distribute them G-vectors or
            frequencies over processes.
        txt: str
            Output file.
        truncation: str or None
            None for no truncation.
            '2D' for standard analytical truncation scheme.
            Non-periodic directions are determined from k-point grid
        eshift: float
            Shift unoccupied bands
        """
        gs, context = get_gs_and_context(calc, txt, world, timer=None)
        wd = get_frequency_descriptor(frequencies, gs=gs, nbands=nbands)

        chi0calc = Chi0Calculator(
            gs, context, nblocks=nblocks,
            wd=wd,
            ecut=ecut, nbands=nbands, eta=eta,
            hilbert=hilbert,
            intraband=intraband,
            disable_point_group=disable_point_group,
            disable_time_reversal=disable_time_reversal,
            integrationmode=integrationmode,
            rate=rate, eshift=eshift
        )
        coulomb = CoulombKernel.from_gs(gs, truncation=truncation)

        super().__init__(chi0calc, coulomb)

    def get_frequencies(self) -> np.ndarray:
        """Return frequencies (in eV) that the χ is evaluated on."""
        return self.chi0calc.wd.omega_w * Hartree

    def get_dynamic_susceptibility(self, *args, xc='ALDA',
                                   filename='chiM_w.csv',
                                   **kwargs):
        dynsus = self.get_inverse_dielectric_function(
            *args, xc=xc, **kwargs).dynamic_susceptibility()
        if filename:
            dynsus.write(filename)
        return dynsus.unpack()

    def get_dielectric_function(self, *args, filename='df.csv', **kwargs):
        """Calculate the dielectric function.

        Generates a file 'df.csv', unless filename is set to None.

        Returns
        -------
        df_NLFC_w: np.ndarray
            Dielectric function without local field corrections.
        df_LFC_w: np.ndarray
            Dielectric functio with local field corrections.
        """
        df = self.get_inverse_dielectric_function(
            *args, **kwargs).macroscopic_dielectric_function()
        if filename:
            df.write(filename)
        return df.unpack()

    def get_eels_spectrum(self, *args, filename='eels.csv', **kwargs):
        """Calculate the macroscopic EELS spectrum.

        Generates a file 'eels.csv', unless filename is set to None.

        Returns
        -------
        eels0_w: np.ndarray
            Spectrum in the independent-particle random-phase approximation.
        eels_w: np.ndarray
            Fully screened EELS spectrum.
        """
        eels = self.get_inverse_dielectric_function(
            *args, **kwargs).eels_spectrum()
        if filename:
            eels.write(filename)
        return eels.unpack()

    def get_polarizability(self, *args, filename='polarizability.csv',
                           **kwargs):
        """Calculate the macroscopic polarizability.

        Generate a file 'polarizability.csv', unless filename is set to None.

        Returns:
        --------
        alpha0_w: np.ndarray
            Polarizability calculated without local-field corrections
        alpha_w: np.ndarray
            Polarizability calculated with local-field corrections.
        """
        pol = self._dielectric_function_new(
            *args, **kwargs).polarizability()
        if filename:
            pol.write(filename)
        return pol.unpack()

    def get_macroscopic_dielectric_constant(self, xc='RPA', direction='x'):
        """Calculate the macroscopic dielectric constant.

        The macroscopic dielectric constant is defined as the real part of the
        dielectric function in the static limit.

        Returns:
        --------
        eps0: float
            Dielectric constant without local field corrections.
        eps: float
            Dielectric constant with local field correction. (RPA, ALDA)
        """
        return self.get_inverse_dielectric_function(
            xc=xc, direction=direction).dielectric_constant()


# ----- Serialized dataclasses and IO ----- #


@dataclass
class ScalarResponseFunctionSet:
    """A set of scalar response functions rf₀(ω) and rf(ω)."""
    wd: FrequencyDescriptor
    rf0_w: np.ndarray
    rf_w: np.ndarray

    @property
    def arrays(self):
        return self.wd.omega_w * Hartree, self.rf0_w, self.rf_w

    def unpack(self):
        # Legacy feature to support old DielectricFunction output format
        # ... to be deprecated ...
        return self.rf0_w, self.rf_w

    def write(self, filename):
        if mpi.rank == 0:
            write_response_function(filename, *self.arrays)

    @property
    def static_limit(self):
        """Return the value of the response functions in the static limit."""
        w0 = np.argmin(np.abs(self.wd.omega_w))
        assert abs(self.wd.omega_w[w0]) < 1e-8
        return np.array([self.rf0_w[w0], self.rf_w[w0]])


def write_response_function(filename, omega_w, rf0_w, rf_w):
    with open(filename, 'w') as fd:
        for omega, rf0, rf in zip(omega_w, rf0_w, rf_w):
            if rf0_w.dtype == complex:
                print('%.6f, %.6f, %.6f, %.6f, %.6f' %
                      (omega, rf0.real, rf0.imag, rf.real, rf.imag),
                      file=fd)
            else:
                print(f'{omega:.6f}, {rf0:.6f}, {rf:.6f}', file=fd)


def read_response_function(filename):
    """Read a stored response function file"""
    d = np.loadtxt(filename, delimiter=',')
    omega_w = np.array(d[:, 0], float)

    if d.shape[1] == 3:
        # Real response function
        rf0_w = np.array(d[:, 1], float)
        rf_w = np.array(d[:, 2], float)
    elif d.shape[1] == 5:
        rf0_w = np.array(d[:, 1], complex)
        rf0_w.imag = d[:, 2]
        rf_w = np.array(d[:, 3], complex)
        rf_w.imag = d[:, 4]
    else:
        raise ValueError(f'Unexpected array dimension {d.shape}')

    return omega_w, rf0_w, rf_w



import pytest
import numpy as np
from ase.build import bulk
from ase.dft.bee import BEEFEnsemble, readbee
from gpaw import GPAW, Mixer, PW
from gpaw.test import gen
from gpaw.mpi import world
import gpaw.cgpaw as cgpaw


@pytest.mark.later
@pytest.mark.mgga
@pytest.mark.libxc
@pytest.mark.slow
@pytest.mark.parametrize('xc', ['mBEEF', 'BEEF-vdW', 'mBEEF-vdW'])
def test_beef(in_tmp_dir, xc):
    if xc[0] == 'm':
        assert cgpaw.lxcXCFuncNum('MGGA_X_MBEEF') is not None

    results = {'mBEEF': (5.449, 0.056),
               'BEEF-vdW': (5.484, 0.071),
               'mBEEF-vdW': (5.426, 0.025)}

    kwargs = dict()
    if xc == 'mBEEF-vdW':
        kwargs['setups'] = dict(Si=gen('Si', xcname='PBEsol'))

    E = []
    V = []
    for a in np.linspace(5.4, 5.5, 3):
        si = bulk('Si', a=a)
        si.calc = GPAW(txt='Si-' + xc + '.txt',
                       mixer=Mixer(0.8, 7, 50.0),
                       xc=xc,
                       kpts=[2, 2, 2],
                       mode=PW(200),
                       **kwargs)
        E.append(si.get_potential_energy())
        ens = BEEFEnsemble(si.calc, verbose=False)
        ens.get_ensemble_energies(200)
        ens.write('Si-{}-{:.3f}'.format(xc, a))
        V.append(si.get_volume())

    p = np.polyfit(V, E, 2)
    v0 = np.roots(np.polyder(p))[0]
    a = (v0 * 4)**(1 / 3)

    a0, da0 = results[xc]

    assert abs(a - a0) < 0.002, (xc, a, a0)

    if world.rank == 0:
        E = []
        for a in np.linspace(5.4, 5.5, 3):
            e = readbee('Si-{}-{:.3f}'.format(xc, a))
            E.append(e)

        A = []
        for energies in np.array(E).T:
            p = np.polyfit(V, energies, 2)
            assert p[0] > 0, (V, E, p)
            v0 = np.roots(np.polyder(p))[0]
            A.append((v0 * 4)**(1 / 3))

        A = np.array(A)
        a = A.mean()
        da = A.std()

        print('a(ref) = {:.3f} +- {:.3f}'.format(a0, da0))
        print('a      = {:.3f} +- {:.3f}'.format(a, da))
        assert abs(a - a0) < 0.01
        assert abs(da - da0) < 0.01



"""
Tools for directmin
"""

import numpy as np
import scipy.linalg as lalg
from copy import deepcopy
from typing import Callable, cast
from gpaw.typing import ArrayND, IntVector, RNG


def expm_ed(a_mat, evalevec=False):
    """
    calculate matrix exponential
    using eigendecomposition of matrix a_mat

    :param a_mat: matrix to be exponented
    :param evalevec: if True then returns eigenvalues
                     and eigenvectors of A

    :return:
    """

    eigval, evec = np.linalg.eigh(1.0j * a_mat)

    product = (evec * np.exp(-1.0j * eigval)) @ evec.T.conj()

    if a_mat.dtype == float:
        product = product.real
    if evalevec:
        return np.ascontiguousarray(product), evec, eigval

    return np.ascontiguousarray(product)


def expm_ed_unit_inv(a_upp_r, oo_vo_blockonly=False):
    """
    calculate matrix exponential using
    Eq. (6) from
    J. Hutter, M. Parrinello, and S. Vogel,
    J. Chem. Phys., 101, 3862 (1994)
    :param a_upp_r: X (see eq in paper)
    :return: unitary matrix
    """
    if np.allclose(a_upp_r, np.zeros_like(a_upp_r)):
        dim_v = a_upp_r.shape[1]
        dim_o = a_upp_r.shape[0]
        if not oo_vo_blockonly:
            dim_v = a_upp_r.shape[1]
            dim_o = a_upp_r.shape[0]

            return np.eye(dim_o + dim_v, dtype=a_upp_r.dtype)
        else:
            return np.vstack([np.eye(dim_o, dtype=a_upp_r.dtype),
                              np.zeros(shape=(dim_v, dim_o),
                                       dtype=a_upp_r.dtype)])

    p_nn = a_upp_r @ a_upp_r.T.conj()
    eigval, evec = np.linalg.eigh(p_nn)
    # Eigenvalues cannot be negative
    eigval[eigval.real < 1.0e-13] = 1.0e-13
    sqrt_eval = np.sqrt(eigval)

    sin_sqrt_p = matrix_function(sqrt_eval, evec, np.sin)
    cos_sqrt_p = matrix_function(sqrt_eval, evec, np.cos)
    sqrt_inv_p = matrix_function(1.0 / sqrt_eval, evec)

    psin = sqrt_inv_p @ sin_sqrt_p
    u_oo = cos_sqrt_p
    u_vo = - a_upp_r.T.conj() @ psin

    if not oo_vo_blockonly:
        inv_p = matrix_function(1.0 / eigval, evec)
        u_ov = psin @ a_upp_r
        dim_v = a_upp_r.shape[1]
        dim_o = a_upp_r.shape[0]
        u_vv = np.eye(dim_v) + \
            a_upp_r.T.conj() @ (cos_sqrt_p - np.eye(dim_o)) @ inv_p @ a_upp_r
        u = np.vstack([
            np.hstack([u_oo, u_ov]),
            np.hstack([u_vo, u_vv])])
    else:
        u = np.vstack([u_oo, u_vo])

    return np.ascontiguousarray(u)


def d_matrix(omega):
    """
    Helper function for calculation of gradient
    w.r.t. skew-hermitian matrix
    see eq. 40 from
    A. V. Ivanov, E. Jónsson, T. Vegge, and H. Jónsso
    Comput. Phys. Commun., 267, 108047 (2021).
    arXiv:2101.12597 [physics.comp-ph]
    """

    m = omega.shape[0]
    u_m = np.ones(shape=(m, m))

    u_m = omega[:, np.newaxis] * u_m - omega * u_m

    with np.errstate(divide='ignore', invalid='ignore'):
        u_m = 1.0j * np.divide(np.exp(-1.0j * u_m) - 1.0, u_m)

    u_m[np.isnan(u_m)] = 1.0
    u_m[np.isinf(u_m)] = 1.0

    return u_m


def minimum_cubic_interpol(x_0, x_1, f_0, f_1, df_0, df_1):
    """
    given f, f' at boundaries of interval [x0, x1]
    calc. x_min where cubic interpolation is minimal
    :return: x_min
    """

    def cubic_function(a, b, c, d, x):
        """
        f(x) = a x^3 + b x^2 + c x + d
        :return: f(x)
        """
        return a * x ** 3 + b * x ** 2 + c * x + d

    if x_0 > x_1:
        x_0, x_1 = x_1, x_0
        f_0, f_1 = f_1, f_0
        df_0, df_1 = df_1, df_0

    r = x_1 - x_0
    a = - 2.0 * (f_1 - f_0) / r ** 3.0 + \
        (df_1 + df_0) / r ** 2.0
    b = 3.0 * (f_1 - f_0) / r ** 2.0 - \
        (df_1 + 2.0 * df_0) / r
    c = df_0
    d = f_0
    D = b ** 2.0 - 3.0 * a * c

    if D < 0.0:
        if f_0 < f_1:
            x_min = x_0
        else:
            x_min = x_1
    else:
        r0 = (-b + np.sqrt(D)) / (3.0 * a) + x_0
        if x_0 < r0 < x_1:
            f_r0 = cubic_function(a, b, c, d, r0 - x_0)
            if f_0 > f_r0 and f_1 > f_r0:
                x_min = r0
            else:
                if f_0 < f_1:
                    x_min = x_0
                else:
                    x_min = x_1
        else:
            if f_0 < f_1:
                x_min = x_0
            else:
                x_min = x_1

    return x_min


def matrix_function(evals, evecs, func=lambda x: x):
    """
    calculate matrix function func(A)
    you need to provide
    :param evals: eigenvalues of A
    :param evecs: eigenvectors of A
    :return: func(A)
    """
    return (evecs * func(evals)) @ evecs.T.conj()


def loewdin_lcao(C_nM, S_MM):
    """
    Loewdin based orthonormalization
    for LCAO mode

    C_nM <- sum_m C_nM[m] [1/sqrt(S)]_mn
    S_mn = (C_nM[m].conj(), S_MM C_nM[n])

    :param C_nM: LCAO coefficients
    :param S_MM: Overlap matrix between basis functions
    :return: Orthonormalized coefficients so that new S_mn = delta_mn
    """

    ev, S_overlapp = np.linalg.eigh(C_nM.conj() @ S_MM @ C_nM.T)
    ev_sqrt = np.diag(1.0 / np.sqrt(ev))

    S = S_overlapp @ ev_sqrt @ S_overlapp.T.conj()

    return S.T @ C_nM


def gramschmidt_lcao(C_nM, S_MM):
    """
    Gram-Schmidt orthonormalization using Cholesky decomposition
    for LCAO mode

    :param C_nM: LCAO coefficients
    :param S_MM: Overlap matrix between basis functions
    :return: Orthonormalized coefficients so that new S_mn = delta_mn
    """

    S_nn = C_nM @ S_MM.conj() @ C_nM.T.conj()
    L_nn = lalg.cholesky(S_nn, lower=True,
                         overwrite_a=True, check_finite=False)
    return lalg.solve(L_nn, C_nM)


def excite(calc, i, a, spin=(0, 0), sort=False):
    """Helper function to initialize a variational excited state calculation.

    Promote an electron from homo + i of k-point spin[0] to lumo + a of
    k-point spin[1].

    Parameters
    ----------
    calc: GPAW instance
        GPAW calculator object.
    i: int
        Subtract 1 from the occupation number of the homo + i orbital of
        k-point spin[0]. E.g. if i=-1, an electron is removed from the
        homo - 1 orbital.
    a: int
        Add 1 to the occupation number of the lumo + a orbital of k-point
        spin[1]. E.g. if a=1, an electron is added to the lumo + 1 orbital.
    spin: tuple of two int
        spin[0] is the k-point from which an electron is removed and spin[1]
        is the k-point where an electron is added.
    sort: bool
        If True, sort the orbitals in the wfs object according to the new
        occupation numbers, and modify the f_n attribute of the kpt objects.
        Default is False.

    Returns
    -------
    list of numpy.ndarray
        List of new occupation numbers. Can be supplied to
        mom.prepare_mom_calculation to initialize an excited state calculation
        with MOM.
    """
    f_sn = [calc.get_occupation_numbers(spin=s).copy()
            for s in range(calc.wfs.nspins)]

    f_n0 = np.asarray(f_sn[spin[0]])
    lumo = len(f_n0[f_n0 > 0])
    homo = lumo - 1

    f_sn[spin[0]][homo + i] -= 1.0
    f_sn[spin[1]][lumo + a] += 1.0

    if sort:
        for s in spin:
            for kpt in calc.wfs.kpt_u:
                if kpt.s == s:
                    kpt.f_n = f_sn[s]
                    changedocc = sort_orbitals_according_to_occ_kpt(
                        calc.wfs, kpt, update_mom=False)[0]
                    if changedocc:
                        f_sn[s] = kpt.f_n

    return f_sn


def sort_orbitals_according_to_occ(
        wfs, constraints=None, update_mom=False, update_eps=True):
    """
    Sort orbitals according to the occupation
    numbers so that there are no holes in the
    distribution of occupation numbers
    :return:
    """
    restart = False
    for kpt in wfs.kpt_u:
        changedocc, ind = sort_orbitals_according_to_occ_kpt(
            wfs, kpt, update_mom=update_mom, update_eps=update_eps)

        if changedocc:
            if constraints:
                k = wfs.kd.nibzkpts * kpt.s + kpt.q
                # Identities of the constrained orbitals have
                # changed and needs to be updated
                constraints[k] = update_constraints_kpt(
                    constraints[k], list(ind))
            restart = True

    return restart


def sort_orbitals_according_to_occ_kpt(
        wfs, kpt, update_mom=False, update_eps=True):
    """
    Sort orbitals according to the occupation
    numbers so that there are no holes in the
    distribution of occupation numbers
    :return:
    """
    changedocc = False
    update_proj = True
    ind = np.array([])

    # Need to initialize the wave functions if
    # restarting from gpw file in fd or pw mode
    if kpt.psit_nG is not None:
        if not isinstance(kpt.psit_nG, np.ndarray):
            wfs.initialize_wave_functions_from_restart_file()
            update_proj = False

    n_occ, occupied = get_n_occ(kpt)
    if n_occ != 0.0 and np.min(kpt.f_n[:n_occ]) == 0:
        ind_occ = np.argwhere(occupied)
        ind_unocc = np.argwhere(~occupied)
        ind = np.vstack((ind_occ, ind_unocc))
        ind = np.squeeze(ind)

        if hasattr(wfs.eigensolver, 'dm_helper'):
            wfs.eigensolver.dm_helper.sort_orbitals(wfs, kpt, ind)
        else:
            sort_orbitals_kpt(wfs, kpt, ind, update_proj)

        kpt.f_n = kpt.f_n[ind]
        if update_eps:
            kpt.eps_n = kpt.eps_n[ind]

        if update_mom:
            # OccupationsMOM.numbers needs
            # to be updated after sorting
            update_mom_numbers(wfs, kpt)

        changedocc = True

    return changedocc, ind


def sort_orbitals_according_to_energies(
        ham, wfs, constraints=None, use_eps=True):
    """
    Sort orbitals according to the eigenvalues or
    the diagonal elements of the Hamiltonian matrix
    """
    if hasattr(wfs.eigensolver, 'dm_helper'):
        dm_helper = wfs.eigensolver.dm_helper
    else:
        dm_helper = None

    for kpt in wfs.kpt_u:
        k = wfs.kd.nibzkpts * kpt.s + kpt.q
        if use_eps:
            orb_energies = kpt.eps_n
        else:
            if dm_helper:
                orb_energies = wfs.eigensolver.dm_helper.orbital_energies(
                    wfs, ham, kpt)
            else:
                raise NotImplementedError

        if dm_helper and dm_helper.func.name == 'PZ-SIC':
            # For SIC, we sort occupied and unoccupied orbitals
            # separately, so the occupation numbers  of canonical
            # and optimal orbitals are always consistent
            n_occ, occupied = get_n_occ(kpt)
            ind_occ = np.argsort(orb_energies[occupied])
            ind_unocc = np.argsort(orb_energies[~occupied])
            ind = np.concatenate((ind_occ, ind_unocc + n_occ))
        else:
            ind = np.argsort(orb_energies)

        # check if it is necessary to sort
        x = np.max(abs(ind - np.arange(len(ind))))
        if x > 0:
            # now sort wfs according to orbital energies
            if dm_helper:
                dm_helper.sort_orbitals(wfs, kpt, ind)
            else:
                sort_orbitals_kpt(wfs, kpt, ind, update_proj=True)

            kpt.f_n[np.arange(len(ind))] = kpt.f_n[ind]

            if use_eps:
                kpt.eps_n[np.arange(len(ind))] = orb_energies[ind]
            else:
                if dm_helper and dm_helper.func.name == 'PZ-SIC':
                    # For SIC, we need to sort both the diagonal elements of
                    # the Lagrange matrix and the self-interaction energies
                    dm_helper.func.lagr_diag_s[k] = orb_energies[ind]
                    dm_helper.func.e_sic_by_orbitals[k] = (
                        dm_helper.func.e_sic_by_orbitals)[k][ind_occ]

            occ_name = getattr(wfs.occupations, "name", None)
            if occ_name == 'mom':
                # OccupationsMOM.numbers needs to be updated
                # after sorting
                update_mom_numbers(wfs, kpt)
            if constraints:
                # Identity if the constrained orbitals have
                # changed and need to be updated
                constraints[k] = update_constraints_kpt(
                    constraints[k], list(ind))


def update_mom_numbers(wfs, kpt):
    if wfs.collinear and wfs.nspins == 1:
        degeneracy = 2
    else:
        degeneracy = 1
    wfs.occupations.numbers[kpt.s] = \
        kpt.f_n / (kpt.weightk * degeneracy)


def sort_orbitals_kpt(wfs, kpt, ind, update_proj=False):
    if wfs.mode == 'lcao':
        kpt.C_nM[np.arange(len(ind)), :] = kpt.C_nM[ind, :]
        wfs.atomic_correction.calculate_projections(wfs, kpt)
    else:
        kpt.psit_nG[np.arange(len(ind))] = kpt.psit_nG[ind]
        if update_proj:
            wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)


def update_constraints_kpt(constraints, ind):
    """
    Change the constraint indices to match a new indexation, e.g. due to
    sorting the orbitals

    :param constraints: The list of constraints for one K-point
    :param ind: List containing information about the change in indexation
    """

    new = deepcopy(constraints)
    for i in range(len(constraints)):
        for k in range(len(constraints[i])):
            new[i][k] = ind.index(constraints[i][k])
    return new


def dict_to_array(x):
    """
    Converts dictionaries with integer keys to one long array by appending.

    :param x: Dictionary
    :return: Long array, dimensions of original dictionary parts, total
             dimensions
    """
    y = []
    dim = []
    dimtot = 0
    for k in x.keys():
        assert isinstance(k, int), (
            'Cannot convert dict to array if keys are not '
            'integer.')
        y += list(x[k])
        dim.append(len(x[k]))
        dimtot += len(x[k])
    return np.asarray(y), dim, dimtot


def array_to_dict(x, dim):
    """
    Converts long array to dictionary with integer keys with values of
    dimensionality specified in dim.

    :param x: Array
    :param dim: List with dimensionalities of parts of the dictionary
    :return: Dictionary
    """
    y = {}
    start = 0
    stop = 0
    for i in range(len(dim)):
        stop += dim[i]
        y[i] = x[start: stop]
        start += dim[i]
    return y


def rotate_orbitals(etdm, wfs, indices, angles, channels):
    """
    Applies rotations between pairs of orbitals.

    :param etdm:       ETDM object for a converged or at least initialized
                       calculation
    :param indices:    List of indices. Each element must be a list of an
                       orbital pair corresponding to the orbital rotation.
                       For occupied-virtual rotations (unitary invariant or
                       sparse representations), the first index represents the
                       occupied, the second the virtual orbital.
                       For occupied-occupied rotations (sparse representation
                       only), the first index must always be smaller than the
                       second.
    :param angles:     List of angles in radians.
    :param channels:   List of spin channels.
    """

    angles = - np.array(angles) * np.pi / 180.0
    a_vec_u = get_a_vec_u(etdm, wfs, indices, angles, channels)
    c = {}
    for kpt in wfs.kpt_u:
        k = etdm.kpointval(kpt)
        c[k] = wfs.kpt_u[k].C_nM.copy()
    etdm.rotate_wavefunctions(wfs, a_vec_u, c)


def get_a_vec_u(etdm, wfs, indices, angles, channels, occ=None):
    """
    Creates an orbital rotation vector based on given indices, angles and
    corresponding spin channels.

    :param etdm:       ETDM object for a converged or at least initialized
                       calculation
    :param indices:    List of indices. Each element must be a list of an
                       orbital pair corresponding to the orbital rotation.
                       For occupied-virtual rotations (unitary invariant or
                       sparse representations), the first index represents the
                       occupied, the second the virtual orbital.
                       For occupied-occupied rotations (sparse representation
                       only), the first index must always be smaller than the
                       second.
    :param angles:     List of angles in radians.
    :param channels:   List of spin channels.
    :param occ:        Occupation numbers for each k-point. Must be specified
                       if the orbitals in the ETDM object are not ordered
                       canonically, as the user orbital indexation is different
                       from the one in the ETDM object then.

    :return new_vec_u: Orbital rotation coordinate vector containing the
                       specified values.
    """

    sort_orbitals_according_to_occ(wfs, etdm.constraints, update_mom=True)

    new_vec_u = {}
    ind_up = etdm.ind_up
    a_vec_u = deepcopy(etdm.a_vec_u)
    conversion = []
    for k in a_vec_u.keys():
        new_vec_u[k] = np.zeros_like(a_vec_u[k])
        if occ is not None:
            f_n = occ[k]
            occupied = f_n > 1.0e-10
            n_occ = len(f_n[occupied])
            if n_occ == 0.0:
                continue
            if np.min(f_n[:n_occ]) == 0:
                ind_occ = np.argwhere(occupied)
                ind_unocc = np.argwhere(~occupied)
                ind = np.vstack((ind_occ, ind_unocc))
                ind = np.squeeze(ind)
                conversion.append(list(ind))
            else:
                conversion.append(None)

    for ind, ang, s in zip(indices, angles, channels):
        if occ is not None:
            if conversion[s] is not None:
                ind[0] = conversion[s].index(ind[0])
                ind[1] = conversion[s].index(ind[1])
        m = np.where(ind_up[s][0] == ind[0])[0]
        n = np.where(ind_up[s][1] == ind[1])[0]
        res = None
        for i in m:
            for j in n:
                if i == j:
                    res = i
        if res is None:
            raise ValueError('Orbital rotation does not exist.')
        new_vec_u[s][res] = ang

    return new_vec_u


def get_n_occ(kpt):
    occupied = kpt.f_n > 1.0e-10
    n_occ = len(kpt.f_n[occupied])

    return n_occ, occupied


def get_indices(dimens):
    return np.tril_indices(dimens, -1)


def random_a(shape, dtype, rng: RNG = cast(RNG, np.random)):
    sample_unit_interval: Callable[[IntVector], ArrayND] = rng.random
    a = sample_unit_interval(shape)
    if dtype == complex:
        a = a.astype(complex)
        a += 1.0j * sample_unit_interval(shape)

    return a



from __future__ import annotations
from typing import Union
from pathlib import Path
from time import ctime
from sys import stdout

from inspect import isgeneratorfunction
from functools import wraps

from ase.utils import IOContext
from ase.utils.timing import Timer

import gpaw.mpi as mpi


TXTFilename = Union[Path, str]


class ResponseContext:
    def __init__(self, txt: TXTFilename = '-',
                 timer=None, comm=mpi.world, mode='w'):
        self.comm = comm
        self.iocontext = IOContext()
        self.open(txt, mode)
        self.set_timer(timer)

    def open(self, txt, mode):
        if txt is stdout and self.comm.rank != 0:
            txt = None
        self.fd = self.iocontext.openfile(txt, self.comm, mode)

    def set_timer(self, timer):
        self.timer = timer or Timer()

    def close(self):
        self.iocontext.close()

    def __del__(self):
        self.close()

    def with_txt(self, txt, mode='w'):
        return ResponseContext(txt=txt, comm=self.comm, timer=self.timer,
                               mode=mode)

    def print(self, *args, flush=True, **kwargs):
        print(*args, file=self.fd, flush=flush, **kwargs)

    def new_txt_and_timer(self, txt, timer=None):
        self.write_timer()
        # Close old output file and create a new
        self.close()
        self.open(txt, mode='w')
        self.set_timer(timer)

    def write_timer(self):
        self.timer.write(self.fd)
        self.print(ctime())


class timer:
    """Decorator for timing a method call.
    NB: Includes copy-paste from ase, which is suboptimal...

    Example::

        from gpaw.response.context import timer

        class A:
            def __init__(self, context):
                self.context = context

            @timer('Add two numbers')
            def add(self, x, y):
                return x + y

        """
    def __init__(self, name):
        self.name = name

    def __call__(self, method):
        if isgeneratorfunction(method):
            @wraps(method)
            def new_method(slf, *args, **kwargs):
                gen = method(slf, *args, **kwargs)
                while True:
                    slf.context.timer.start(self.name)
                    try:
                        x = next(gen)
                    except StopIteration:
                        break
                    finally:
                        slf.context.timer.stop()
                    yield x
        else:
            @wraps(method)
            def new_method(slf, *args, **kwargs):
                slf.context.timer.start(self.name)
                x = method(slf, *args, **kwargs)
                try:
                    slf.context.timer.stop()
                except IndexError:
                    pass
                return x
        return new_method



import numpy as np
from gpaw.core.matrix import Matrix
from gpaw.lcao.tci import TCIExpansions
from gpaw.new import zips
from gpaw.new.fd.builder import FDDFTComponentsBuilder
from gpaw.new.lcao.ibzwfs import LCAOIBZWaveFunctions
from gpaw.new.lcao.eigensolver import LCAOEigensolver
from gpaw.new.lcao.forces import TCIDerivatives
from gpaw.new.lcao.hamiltonian import LCAOHamiltonian
from gpaw.new.lcao.hybrids import HybridLCAOEigensolver, HybridXCFunctional
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions
from gpaw.utilities.timing import NullTimer


class LCAODFTComponentsBuilder(FDDFTComponentsBuilder):
    def __init__(self,
                 atoms,
                 params,
                 *,
                 comm,
                 distribution=None):
        super().__init__(atoms, params, comm=comm)
        self.distribution = distribution
        self.basis = None

    def create_wf_description(self):
        raise NotImplementedError

    def create_xc_functional(self):
        if self.params.xc['name'] in ['HSE06', 'PBE0', 'EXX']:
            return HybridXCFunctional(self.params.xc)
        return super().create_xc_functional()

    def create_basis_set(self):
        self.basis = FDDFTComponentsBuilder.create_basis_set(self)
        return self.basis

    def create_hamiltonian_operator(self):
        return LCAOHamiltonian(self.basis)

    def create_eigensolver(self, hamiltonian):
        if self.params.xc['name'] in ['HSE06', 'PBE0', 'EXX']:
            return HybridLCAOEigensolver(self.basis,
                                         self.fracpos_ac,
                                         self.grid.cell_cv)
        return LCAOEigensolver(self.basis)

    def read_ibz_wave_functions(self, reader):
        c = 1
        if reader.version >= 0 and reader.version < 4:
            c = reader.bohr**1.5

        basis = self.create_basis_set()
        potential = self.create_potential_calculator()
        if 'coefficients' in reader.wave_functions:
            coefficients = reader.wave_functions.proxy('coefficients')
            coefficients.scale = c
        else:
            coefficients = None

        ibzwfs = self.create_ibz_wave_functions(basis, potential,
                                                coefficients=coefficients)

        # Set eigenvalues, occupations, etc..
        self.read_wavefunction_values(reader, ibzwfs)
        return ibzwfs

    def create_ibz_wave_functions(self,
                                  basis,
                                  potential,
                                  *,
                                  log=None,
                                  coefficients=None):
        ibzwfs, _ = create_lcao_ibzwfs(
            basis,
            self.ibz, self.communicators, self.setups,
            self.fracpos_ac, self.grid, self.dtype,
            self.nbands, self.ncomponents, self.atomdist, self.nelectrons,
            coefficients)
        return ibzwfs


def create_lcao_ibzwfs(basis,
                       ibz, communicators, setups,
                       fracpos_ac, grid, dtype,
                       nbands, ncomponents, atomdist, nelectrons,
                       coefficients=None):
    kpt_band_comm = communicators['D']
    kpt_comm = communicators['k']
    band_comm = communicators['b']
    domain_comm = communicators['d']

    S_qMM, T_qMM, P_qaMi, tciexpansions, tci_derivatives = tci_helper(
        basis, ibz, domain_comm, band_comm, kpt_comm,
        fracpos_ac, atomdist,
        grid, dtype, setups)

    nao = setups.nao

    def create_wfs(spin, q, k, kpt_c, weight):
        C_nM = Matrix(nbands, 2 * nao if ncomponents == 4 else nao,
                      dtype,
                      dist=(band_comm, band_comm.size, 1))
        if coefficients is not None:
            C_nM.data[:] = coefficients.proxy(spin, k)
        return LCAOWaveFunctions(
            setups=setups,
            tci_derivatives=tci_derivatives,
            basis=basis,
            C_nM=C_nM,
            S_MM=S_qMM[q],
            T_MM=T_qMM[q],
            P_aMi=P_qaMi[q],
            kpt_c=kpt_c,
            fracpos_ac=fracpos_ac,
            atomdist=atomdist,
            domain_comm=domain_comm,
            spin=spin,
            q=q,
            k=k,
            weight=weight,
            ncomponents=ncomponents)

    ibzwfs = LCAOIBZWaveFunctions.create(
        ibz=ibz,
        nelectrons=nelectrons,
        ncomponents=ncomponents,
        create_wfs_func=create_wfs,
        kpt_comm=kpt_comm,
        kpt_band_comm=kpt_band_comm,
        comm=communicators['w'])
    ibzwfs.grid = grid  # The TCI-stuff needs cell and pbc from somewhere ...
    return ibzwfs, tciexpansions


def tci_helper(basis,
               ibz,
               domain_comm, band_comm, kpt_comm,
               fracpos_ac, atomdist,
               grid,
               dtype,
               setups):
    rank_k = ibz.ranks(kpt_comm)
    here_k = rank_k == kpt_comm.rank
    kpt_qc = ibz.kpt_kc[here_k]

    tciexpansions = TCIExpansions.new_from_setups(setups)
    manytci = tciexpansions.get_manytci_calculator(
        setups, grid._gd, fracpos_ac,
        kpt_qc, dtype, NullTimer())

    my_atom_indices = basis.my_atom_indices
    M1 = basis.Mstart
    M2 = basis.Mstop
    S0_qMM, T0_qMM = manytci.O_qMM_T_qMM(domain_comm, M1, M2, True)
    if dtype == complex:
        np.negative(S0_qMM.imag, S0_qMM.imag)
        np.negative(T0_qMM.imag, T0_qMM.imag)

    P_aqMi = manytci.P_aqMi(my_atom_indices)
    P_qaMi = [{a: P_aqMi[a][q] for a in my_atom_indices}
              for q in range(len(S0_qMM))]

    for a, P_qMi in P_aqMi.items():
        dO_ii = setups[a].dO_ii
        for P_Mi, S_MM in zips(P_qMi, S0_qMM):
            S_MM += P_Mi[M1:M2].conj() @ dO_ii @ P_Mi.T
    domain_comm.sum(S0_qMM)

    # self.atomic_correction= self.atomic_correction_cls.new_from_wfs(self)
    # self.atomic_correction.add_overlap_correction(newS_qMM)

    nao = setups.nao

    S_qMM = [Matrix(nao, nao, data=S_MM,
                    dist=(band_comm, band_comm.size, 1)) for S_MM in S0_qMM]
    T_qMM = [Matrix(nao, nao, data=T_MM,
                    dist=(band_comm, band_comm.size, 1)) for T_MM in T0_qMM]

    for S_MM in S_qMM:
        S_MM.tril2full()
    for T_MM in T_qMM:
        T_MM.tril2full()

    tci_derivatives = TCIDerivatives(manytci, atomdist, nao)

    return S_qMM, T_qMM, P_qaMi, tciexpansions, tci_derivatives



"""
Potentials for orbital density dependent energy functionals
"""

import numpy as np

from gpaw.directmin.tools import d_matrix
from gpaw.utilities import unpack_hermitian


class KSFDPW:
    """
    KS-DFT
    """
    def __init__(self, wfs, dens, ham):

        self.name = 'Zero'
        self.n_kps = wfs.kd.nibzkpts
        self.grad = {}
        self.total_sic = 0.0
        self.eks = 0.0
        self.changedocc = 0
        self.dens = dens
        self.ham = ham

    def get_energy_and_gradients_kpt(
            self, wfs, kpt, grad_knG=None, U_k=None, add_grad=False, ham=None):

        k = self.n_kps * kpt.s + kpt.q
        nbands = wfs.bd.nbands
        if U_k is not None:
            assert U_k[k].shape[0] == nbands

        wfs.timer.start('e/g grid calculations')
        self.grad[k] = wfs.empty(nbands, q=kpt.q)
        wfs.apply_pseudo_hamiltonian(kpt, ham, kpt.psit_nG, self.grad[k])

        c_axi = {}
        for a, P_xi in kpt.P_ani.items():
            dH_ii = unpack_hermitian(ham.dH_asp[a][kpt.s])
            c_xi = np.dot(P_xi, dH_ii)
            c_axi[a] = c_xi

        # not sure about this:
        ham.xc.add_correction(kpt, kpt.psit_nG, self.grad[k],
                              kpt.P_ani, c_axi, n_x=None,
                              calculate_change=False)
        # add projectors to the H|psi_i>
        wfs.pt.add(self.grad[k], c_axi, kpt.q)
        # scale with occupation numbers
        for i, f in enumerate(kpt.f_n):
            self.grad[k][i] *= f

        if add_grad:
            if U_k is not None:
                grad_knG[k] += \
                    np.tensordot(U_k[k].conj(), self.grad[k], axes=1)
            else:
                grad_knG[k] += self.grad[k]
        else:
            if U_k is not None:
                self.grad[k][:] = np.tensordot(U_k[k].conj(),
                                               self.grad[k], axes=1)

        wfs.timer.stop('e/g grid calculations')

        return 0.0

    def get_energy_and_gradients_inner_loop(
            self, wfs, kpt, a_mat, evals, evec, ham=None, exstate=True):

        if not exstate:
            raise RuntimeError('Attempting to optimize unitary-invariant '
                               'energy in occupied-occupied rotation space.')

        ndim = wfs.bd.nbands

        k = self.n_kps * kpt.s + kpt.q
        self.grad[k] = np.zeros_like(kpt.psit_nG[:ndim])
        e_sic = self.get_energy_and_gradients_kpt(wfs, kpt, ham=ham)
        wfs.timer.start('Unitary gradients')
        l_odd = wfs.integrate(kpt.psit_nG[:ndim], self.grad[k][:ndim], True)
        f = np.ones(ndim)
        indz = np.absolute(l_odd) > 1.0e-4
        l_c = 2.0 * l_odd[indz]
        l_odd = f[:, np.newaxis] * l_odd.T.conj() - f * l_odd
        kappa = np.max(np.absolute(l_odd[indz]) / np.absolute(l_c))

        if a_mat is None:
            wfs.timer.stop('Unitary gradients')
            return 2.0 * l_odd.T.conj(), e_sic, kappa
        else:
            g_mat = evec.T.conj() @ l_odd.T.conj() @ evec
            g_mat = g_mat * d_matrix(evals)
            g_mat = evec @ g_mat @ evec.T.conj()
            for i in range(g_mat.shape[0]):
                g_mat[i][i] *= 0.5
            wfs.timer.stop('Unitary gradients')
            if a_mat.dtype == float:
                g_mat = g_mat.real
            return 2.0 * g_mat, e_sic, kappa



from math import pi
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.density import Density


class LCAOIBZWaveFunctions(IBZWaveFunctions):
    def move(self, fracpos_ac, atomdist):
        from gpaw.new.lcao.builder import tci_helper

        super().move(fracpos_ac, atomdist)

        for wfs in self:
            basis = wfs.basis
            setups = wfs.setups
            break
        basis.set_positions(fracpos_ac)
        myM = (basis.Mmax + self.band_comm.size - 1) // self.band_comm.size
        basis.set_matrix_distribution(
            min(self.band_comm.rank * myM, basis.Mmax),
            min((self.band_comm.rank + 1) * myM, basis.Mmax))

        S_qMM, T_qMM, P_qaMi, tciexpansions, tci_derivatives = tci_helper(
            basis, self.ibz, self.domain_comm, self.band_comm, self.kpt_comm,
            fracpos_ac, atomdist,
            self.grid, self.dtype, setups)

        for wfs in self:
            wfs.tci_derivatives = tci_derivatives
            wfs.S_MM = S_qMM[wfs.q]
            wfs.T_MM = T_qMM[wfs.q]
            wfs.P_aMi = P_qaMi[wfs.q]

    def normalize_density(self, density: Density) -> None:
        """Normalize density.

        Basis functions may extend outside box!
        """
        pseudo_charge = density.nt_sR.integrate().sum()
        ccc_aL = density.calculate_compensation_charge_coefficients()
        comp_charge = (4 * pi)**0.5 * sum(float(ccc_L[0])
                                          for ccc_L in ccc_aL.values())
        comp_charge = ccc_aL.layout.atomdist.comm.sum_scalar(comp_charge)
        density.nt_sR.data *= -comp_charge / pseudo_charge



import numpy as np
from ase.units import Hartree

import gpaw.cgpaw as cgpaw
from gpaw.xc import XC
from gpaw.xc.kernel import XCKernel
from gpaw.xc.libxc import LibXC
from gpaw.xc.vdw import VDWFunctional
from gpaw import debug


class BEE2(XCKernel):
    """GGA exchange expanded in Legendre polynomials."""
    def __init__(self, parameters=None):
        """BEE2.

        parameters: array
            [transformation,0.0,[orders],[coefs]].

        """

        if parameters is None:
            parameters = (
                [4.0, 0.0] + list(range(30)) +
                [1.516501714304992365356, 0.441353209874497942611,
                 -0.091821352411060291887, -0.023527543314744041314,
                 0.034188284548603550816, 0.002411870075717384172,
                 -0.014163813515916020766, 0.000697589558149178113,
                 0.009859205136982565273, -0.006737855050935187551,
                 -0.001573330824338589097, 0.005036146253345903309,
                 -0.002569472452841069059, -0.000987495397608761146,
                 0.002033722894696920677, -0.000801871884834044583,
                 -0.000668807872347525591, 0.001030936331268264214,
                 -0.000367383865990214423, -0.000421363539352619543,
                 0.000576160799160517858, -0.000083465037349510408,
                 -0.000445844758523195788, 0.000460129009232047457,
                 -0.000005231775398304339, -0.000423957047149510404,
                 0.000375019067938866537, 0.000021149381251344578,
                 -0.000190491156503997170, 0.000073843624209823442])
        else:
            assert len(parameters) > 2
            assert np.mod(len(parameters), 2) == 0
            assert parameters[1] == 0.0

        parameters = np.array(parameters, dtype=float).ravel()
        self.xc = cgpaw.XCFunctional(17, parameters)
        self.type = 'GGA'
        self.name = 'BEE2'


class BEEVDWKernel(XCKernel):
    """Kernel for BEEVDW functionals."""
    def __init__(self, bee, xcoefs, ldac, ggac):
        """BEEVDW kernel.

        parameters:

        bee : str
            choose BEE1 or BEE2 exchange basis expansion.
        xcoefs : array
            coefficients for exchange.
        ldac : float
            coefficient for LDA correlation.
        pbec : float
            coefficient for PBE correlation.

        """

        if bee == 'BEE2':
            self.BEE = BEE2(xcoefs)
            self.GGAc = LibXC('GGA_C_PBE')
            self.xtype = 'GGA'
            self.type = 'GGA'
        elif bee == 'BEE3':
            self.BEE = LibXC('MGGA_X_MBEEFVDW')
            self.GGAc = LibXC('GGA_C_PBE_SOL')
            self.xtype = 'MGGA'
            self.type = 'MGGA'
        else:
            raise ValueError('Unknown BEE exchange: %s', bee)

        self.LDAc = LibXC('LDA_C_PW')
        self.ldac = ldac
        self.ggac = ggac
        if bee in ['BEE1', 'BEE2']:
            self.ggac -= 1.0
        self.name = 'BEEVDW'

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg=None, dedsigma_xg=None,
                  tau_sg=None, dedtau_sg=None):
        if debug:
            self.check_arguments(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                                 tau_sg, dedtau_sg)

        if self.xtype == 'GGA':
            self.BEE.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        elif self.xtype == 'MGGA':
            self.BEE.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                               tau_sg, dedtau_sg)
        else:
            raise ValueError('Unexpected value of xtype:', self.xtype)

        e0_g = np.empty_like(e_g)
        dedn0_sg = np.empty_like(dedn_sg)
        dedsigma0_xg = np.empty_like(dedsigma_xg)
        for coef, kernel in [(self.ldac, self.LDAc),
                             (self.ggac, self.GGAc)]:
            dedn0_sg[:] = 0.0
            kernel.calculate(e0_g, n_sg, dedn0_sg, sigma_xg, dedsigma0_xg)
            e_g += coef * e0_g
            dedn_sg += coef * dedn0_sg
            if kernel.type == 'GGA':
                dedsigma_xg += coef * dedsigma0_xg


class BEEFEnsemble:
    """BEEF ensemble error estimation."""
    def __init__(self, calc):
        """BEEF ensemble

        parameters:

        calc: object
            Calculator holding a selfconsistent BEEF type electron density.
            May be BEEF-vdW or mBEEF.
        """

        self.calc = calc

        self.e_dft = None
        self.e0 = None

        # determine functional and read parameters
        self.xc = self.calc.get_xc_functional()
        if self.xc == 'BEEF-vdW':
            self.bee_type = 1
        elif self.xc == 'mBEEF':
            self.bee_type = 2
            self.max_order = 8
            self.trans = [6.5124, -1.0]
            self.calc.converge_wave_functions()
            self.calc.log('wave functions converged')
        elif self.xc == 'mBEEF-vdW':
            self.bee_type = 3
            self.max_order = 5
            self.trans = [6.5124, -1.0]
            self.calc.converge_wave_functions()
            self.calc.log('wave functions converged')
        else:
            raise NotImplementedError('xc = %s not implemented' % self.xc)

    def create_xc_contributions(self, type):
        """General function for creating exchange or correlation energies"""
        assert type in ['exch', 'corr']
        err = 'bee_type %i not implemented' % self.bee_type

        if type == 'exch':
            if self.bee_type == 1:
                out = self.beefvdw_energy_contribs_x()
            elif self.bee_type in [2, 3]:
                out = self.mbeef_exchange_energy_contribs()
            else:
                raise NotImplementedError(err)
        else:
            if self.bee_type == 1:
                out = self.beefvdw_energy_contribs_c()
            elif self.bee_type == 2:
                out = np.array([])
            elif self.bee_type == 3:
                out = self.mbeefvdw_energy_contribs_c()
            else:
                raise NotImplementedError(err)
        return out

    def get_non_xc_total_energies(self):
        """Compile non-XC total energy contributions"""
        if self.e_dft is None:
            self.e_dft = self.calc.get_potential_energy()
        if self.e0 is None:
            from gpaw.xc.kernel import XCNull
            xc_null = XC(XCNull())
            self.e0 = self.e_dft + self.calc.get_xc_difference(xc_null)
        assert isinstance(self.e_dft, float)
        assert isinstance(self.e0, float)

    def mbeef_exchange_energy_contribs(self):
        """Legendre polynomial exchange contributions to mBEEF Etot"""
        self.get_non_xc_total_energies()
        e_x = np.zeros((self.max_order, self.max_order))
        for p1 in range(self.max_order):  # alpha
            for p2 in range(self.max_order):  # s2
                pars_i = np.array([1, self.trans[0], p2, 1.0])
                pars_j = np.array([1, self.trans[1], p1, 1.0])
                pars = np.hstack((pars_i, pars_j))
                x = XC('2D-MGGA', pars)
                e_x[p1, p2] = (self.e_dft +
                               self.calc.get_xc_difference(x) - self.e0)
                del x
        return e_x

    def beefvdw_energy_contribs_x(self):
        """Legendre polynomial exchange contributions to BEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_pbe = (self.e_dft + self.calc.get_xc_difference('GGA_C_PBE') -
                 self.e0)

        exch = np.zeros(30)
        for p in range(30):
            pars = [4, 0, p, 1.0]
            bee = XC('BEE2', pars)
            exch[p] = (self.e_dft + self.calc.get_xc_difference(bee) -
                       self.e0 - e_pbe)
            del bee
        return exch

    def beefvdw_energy_contribs_c(self):
        """LDA and PBE correlation contributions to BEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_lda = self.e_dft + self.calc.get_xc_difference('LDA_C_PW') - self.e0
        e_pbe = self.e_dft + self.calc.get_xc_difference('GGA_C_PBE') - self.e0
        corr = np.array([e_lda, e_pbe])
        return corr

    def mbeefvdw_energy_contribs_c(self):
        """LDA, PBEsol, and nl2 correlation contributions to mBEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_lda = self.e_dft + self.calc.get_xc_difference('LDA_C_PW') - self.e0
        e_sol = (self.e_dft + self.calc.get_xc_difference('GGA_C_PBE_SOL') -
                 self.e0)
        vdwdf2 = VDWFunctional('vdW-DF2')
        self.calc.get_xc_difference(vdwdf2)
        e_nl2 = vdwdf2.get_Ecnl() * Hartree
        corr = np.array([e_lda, e_sol, e_nl2])
        return corr



from functools import partial

import numpy as np
from ase.utils.timing import timer
from gpaw import debug
from gpaw.eigensolvers.diagonalizerbackend import (ScalapackDiagonalizer,
                                                   ScipyDiagonalizer)
from gpaw.eigensolvers.eigensolver import Eigensolver
from gpaw.hybrids import HybridXC
from gpaw.matrix import matrix_matrix_multiply as mmm


class DummyArray:
    def __getitem__(self, x):
        return np.empty((0, 0))


class Davidson(Eigensolver):
    """Simple Davidson eigensolver

    It is expected that the trial wave functions are orthonormal
    and the integrals of projector functions and wave functions
    ``nucleus.P_uni`` are already calculated.

    Solution steps are:

    * Subspace diagonalization
    * Calculate all residuals
    * Add preconditioned residuals to the subspace and diagonalize
    """

    def __init__(
            self, niter=2):
        Eigensolver.__init__(self)
        self.niter = niter
        self.diagonalizer_backend = None

        self.orthonormalization_required = False
        self.H_NN = DummyArray()
        self.S_NN = DummyArray()
        self.eps_N = DummyArray()

    def __repr__(self):
        return 'Davidson(niter=%d)' % (
            self.niter)

    def todict(self):
        return {'name': 'dav', 'niter': self.niter}

    def initialize(self, wfs):
        Eigensolver.initialize(self, wfs)
        slcomm, nrows, ncols, slsize = wfs.scalapack_parameters

        if wfs.gd.comm.rank == 0 and wfs.bd.comm.rank == 0:
            # Allocate arrays
            B = self.nbands
            self.H_NN = np.zeros((2 * B, 2 * B), self.dtype)
            self.S_NN = np.zeros((2 * B, 2 * B), self.dtype)
            self.eps_N = np.zeros(2 * B)

        if slsize is not None:
            self.diagonalizer_backend = ScalapackDiagonalizer(
                arraysize=self.nbands * 2,
                grid_nrows=nrows,
                grid_ncols=ncols,
                scalapack_communicator=slcomm,
                dtype=self.dtype,
                blocksize=slsize)
        else:
            self.diagonalizer_backend = ScipyDiagonalizer(slcomm)

    def estimate_memory(self, mem, wfs):
        Eigensolver.estimate_memory(self, mem, wfs)
        nbands = wfs.bd.nbands
        mem.subnode('H_nn', nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('S_nn', nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('H_2n2n', 4 * nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('S_2n2n', 4 * nbands * nbands * mem.itemsize[wfs.dtype])
        mem.subnode('eps_2n', 2 * nbands * mem.floatsize)

    @timer('Davidson')
    def iterate_one_k_point(self, ham, wfs, kpt, weights):
        """Do Davidson iterations for the kpoint"""
        if isinstance(ham.xc, HybridXC):
            self.niter = 1

        bd = wfs.bd
        B = bd.nbands

        H_NN = self.H_NN
        S_NN = self.S_NN
        eps_N = self.eps_N

        def integrate(a_G):
            if wfs.collinear:
                return np.real(wfs.integrate(a_G, a_G, global_integral=False))
            return sum(
                np.real(wfs.integrate(b_G, b_G, global_integral=False))
                for b_G in a_G)

        self.subspace_diagonalize(ham, wfs, kpt)

        psit = kpt.psit
        psit2 = psit.new(buf=wfs.work_array)
        P = kpt.projections
        P2 = P.new()
        P3 = P.new()
        M = wfs.work_matrix_nn
        dS = wfs.setups.dS
        comm = wfs.gd.comm

        is_gridband_master = (comm.rank == 0) and (bd.comm.rank == 0)

        if bd.comm.size > 1:
            M0 = M.new(dist=(bd.comm, 1, 1))
        else:
            M0 = M

        if comm.rank == 0:
            e_N = bd.collect(kpt.eps_n)
            if e_N is not None:
                eps_N[:B] = e_N

        Ht = partial(wfs.apply_pseudo_hamiltonian, kpt, ham)

        if self.keep_htpsit:
            R = psit.new(buf=self.Htpsit_nG)
        else:
            R = psit.apply(Ht)

        self.calculate_residuals(kpt, wfs, ham, psit, P, kpt.eps_n, R, P2)

        precond = self.preconditioner

        for nit in range(self.niter):
            if nit == self.niter - 1:
                error = np.dot(weights, [integrate(R_G) for R_G in R.array])

            for psit_G, R_G, psit2_G in zip(psit.array, R.array, psit2.array):
                ekin = precond.calculate_kinetic_energy(psit_G, kpt)
                precond(R_G, kpt, ekin, out=psit2_G)

            # Calculate projections
            psit2.matrix_elements(wfs.pt, out=P2)

            def copy(M, C_nn):
                comm.sum(M.array, 0)
                if comm.rank == 0:
                    M.complex_conjugate()
                    M.redist(M0)
                    if bd.comm.rank == 0:
                        C_nn[:] = M0.array

            with self.timer('calc. matrices'):
                # <psi2 | H | psi2>
                psit2.matrix_elements(
                    operator=Ht, result=R, out=M, symmetric=True, cc=True)
                ham.dH(P2, out=P3)
                mmm(1.0, P2, 'N', P3, 'C', 1.0, M)  # , symmetric=True)
                copy(M, H_NN[B:, B:])

                # <psi2 | H | psi>
                R.matrix_elements(psit, out=M, cc=True)
                mmm(1.0, P3, 'N', P, 'C', 1.0, M)
                copy(M, H_NN[B:, :B])

                # <psi2 | S | psi2>
                psit2.matrix_elements(out=M, symmetric=True, cc=True)
                dS.apply(P2, out=P3)
                mmm(1.0, P2, 'N', P3, 'C', 1.0, M)
                copy(M, S_NN[B:, B:])

                # <psi2 | S | psi>
                psit2.matrix_elements(psit, out=M, cc=True)
                mmm(1.0, P3, 'N', P, 'C', 1.0, M)
                copy(M, S_NN[B:, :B])

            if is_gridband_master:
                H_NN[:B, :B] = np.diag(eps_N[:B])
                S_NN[:B, :B] = np.eye(B)

            with self.timer('diagonalize'):
                if is_gridband_master and debug:
                    H_NN[np.triu_indices(2 * B, 1)] = 42.0
                    S_NN[np.triu_indices(2 * B, 1)] = 42.0

                try:
                    self.diagonalizer_backend.diagonalize(
                        H_NN, S_NN, eps_N, debug=debug)
                except np.linalg.LinAlgError as ex:
                    raise ValueError(
                        'Too few plane waves or grid points') from ex
            if comm.rank == 0:
                bd.distribute(eps_N[:B], kpt.eps_n)
            comm.broadcast(kpt.eps_n, 0)

            with self.timer('rotate_psi'):
                if comm.rank == 0:
                    if bd.comm.rank == 0:
                        M0.array[:] = H_NN[:B, :B].T
                    M0.redist(M)
                comm.broadcast(M.array, 0)
                mmm(1.0, M, 'N', psit, 'N', 0.0, R)
                mmm(1.0, M, 'N', P, 'N', 0.0, P3)
                if comm.rank == 0:
                    if bd.comm.rank == 0:
                        M0.array[:] = H_NN[B:, :B].T
                    M0.redist(M)
                comm.broadcast(M.array, 0)
                mmm(1.0, M, 'N', psit2, 'N', 1.0, R)
                mmm(1.0, M, 'N', P2, 'N', 1.0, P3)
                psit[:] = R
                P, P3 = P3, P
                kpt.projections = P

            if nit < self.niter - 1:
                psit.apply(Ht, out=R)
                self.calculate_residuals(
                    kpt, wfs, ham, psit, P, kpt.eps_n, R, P2)

        error = wfs.gd.comm.sum_scalar(error)
        return error



"""Test Hirshfeld for spin/no spin consistency."""
import pytest
from ase import Atoms
from ase.parallel import parprint

from gpaw import GPAW, FermiDirac
from gpaw.analyse.hirshfeld import HirshfeldPartitioning


@pytest.mark.later
def test_vdw_H_Hirshfeld():
    h = 0.25
    box = 3

    atoms = Atoms('H')
    atoms.center(vacuum=box)

    volumes = []
    for spinpol in [False, True]:
        calc = GPAW(mode='fd',
                    h=h,
                    occupations=FermiDirac(0.1, fixmagmom=spinpol),
                    spinpol=spinpol)
        calc.calculate(atoms)
        vol = HirshfeldPartitioning(calc).get_effective_volume_ratios()
        volumes.append(vol)
    parprint(volumes)
    assert volumes[0][0] == pytest.approx(volumes[1][0], abs=4e-9)



from gpaw.xc import xc_string_to_dict
from ase.utils import basestring
from gpaw.directmin.functional.fdpw.ks import KSFDPW
from gpaw.directmin.functional.fdpw.pz import PZSICFDPW


def get_functional(func, *args):

    if isinstance(func, KSFDPW) or isinstance(func, PZSICFDPW):
        return func
    elif isinstance(func, basestring):
        func = xc_string_to_dict(func)

    if isinstance(func, dict):
        kwargs = func.copy()
        name = kwargs.pop('name').replace('-', '').lower()
        functional = {'ks': KSFDPW,
                      'pzsic': PZSICFDPW}[name](*args, **kwargs)
        return functional
    else:
        raise TypeError('Check functional parameter.')



import numpy as np
from numpy import sqrt, pi, exp, abs
from scipy.special import erf

import gpaw.cgpaw as cgpaw
from gpaw import debug
from gpaw.utilities.tools import coordinates
from gpaw.utilities import is_contiguous


def Y_L(L, x, y, z, r2):
    if L == 0:
        return 0.28209479177387814
    elif L == 1:
        return 0.48860251190291992 * y
    elif L == 2:
        return 0.48860251190291992 * z
    elif L == 3:
        return 0.48860251190291992 * x
    elif L == 4:
        return 1.0925484305920792 * x * y
    elif L == 5:
        return 1.0925484305920792 * y * z
    elif L == 6:
        return 0.31539156525252005 * (3 * z * z - r2)
    elif L == 7:
        return 1.0925484305920792 * x * z
    elif L == 8:
        return 0.54627421529603959 * (x * x - y * y)


def gauss_L(a, L, x, y, z, r2, exp_ar2):
    if L == 0:
        return sqrt(a**3 * 4) / pi * exp_ar2
    elif L == 1:
        return sqrt(a**5 * 5.333333333333333) / pi * y * exp_ar2
    elif L == 2:
        return sqrt(a**5 * 5.333333333333333) / pi * z * exp_ar2
    elif L == 3:
        return sqrt(a**5 * 5.333333333333333) / pi * x * exp_ar2
    elif L == 4:
        return sqrt(a**7 * 4.2666666666666666) / pi * x * y * exp_ar2
    elif L == 5:
        return sqrt(a**7 * 4.2666666666666666) / pi * y * z * exp_ar2
    elif L == 6:
        return sqrt(
            a**7 * 0.35555555555555557) / pi * (3 * z * z - r2) * exp_ar2
    elif L == 7:
        return sqrt(a**7 * 4.2666666666666666) / pi * x * z * exp_ar2
    elif L == 8:
        return sqrt(a**7 * 1.0666666666666667) / pi * (x * x - y * y) * exp_ar2


def gausspot_L(a, L, x, y, z, r, r2, erf_sar, exp_ar2):
    if L == 0:
        return 2.0 * 1.7724538509055159 * erf_sar / r
    elif L == 1:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * y
    elif L == 2:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * z
    elif L == 3:
        return 1.1547005383792515 * (1.7724538509055159 * erf_sar -
                                     2 * sqrt(a) * r * exp_ar2) / r / r2 * x
    elif L == 4:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * x * y
    elif L == 5:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * y * z
    elif L == 6:
        return 0.14907119849998599 * (5.3173615527165481 * erf_sar -
                                      (6 + 4 *
                                       (sqrt(a) * r)**2) * sqrt(a) * r *
                                      exp_ar2) / r / r2**2 * (3 * z * z - r2)
    elif L == 7:
        return 0.5163977794943222 * (
            5.3173615527165481 * erf_sar -
            (6 + 4 *
             (sqrt(a) * r)**2) * sqrt(a) * r * exp_ar2) / r / r2**2 * x * z
    elif L == 8:
        return 0.2581988897471611 * (5.3173615527165481 * erf_sar -
                                     (6 + 4 * (sqrt(a) * r)**2) * sqrt(a) * r *
                                     exp_ar2) / r / r2**2 * (x * x - y * y)


# end of computer generated code


class Gaussian:
    r"""Class offering several utilities related to the generalized gaussians.

    Generalized gaussians are defined by::

                       _____                           2
                      /  1       l!         l+3/2  -a r   l  m
       g (x,y,z) =   / ----- --------- (4 a)      e      r  Y (x,y,z),
        L          \/  4 pi  (2l + 1)!                       l

    where a is the inverse width of the gaussian, and Y_l^m is a real
    spherical harmonic.
    The gaussians are centered in the middle of input grid-descriptor."""
    def __init__(self, gd, a=19., center=None):
        self.gd = gd
        self.xyz, self.r2 = coordinates(gd, center)
        self.r = np.sqrt(self.r2)
        self.set_width(a, center)
        self.exp_ar2 = exp(-self.a * self.r2)
        self.erf_sar = erf(sqrt(self.a) * self.r)

    def set_width(self, a, center):
        """Set exponent of exp-function to -a on the boundary."""
        if center is None:
            self.a = 4 * a * (self.gd.icell_cv**2).sum(1).max()
        else:
            cell_center = self.gd.cell_cv.sum(1) / 2
            r_min = (cell_center - np.abs(center - cell_center)).min()
            self.a = a / r_min**2

    def get_gauss(self, L):
        a = self.a
        x, y, z = tuple(self.xyz)
        r2 = self.r2
        exp_ar2 = self.exp_ar2
        return gauss_L(a, L, x, y, z, r2, exp_ar2)

    def get_gauss_pot(self, L):
        a = self.a
        x, y, z = tuple(self.xyz)
        r2 = self.r2
        r = self.r
        erf_sar = self.erf_sar
        exp_ar2 = self.exp_ar2
        return gausspot_L(a, L, x, y, z, r, r2, erf_sar, exp_ar2)

    def get_moment(self, n, L):
        r2 = self.r2
        x, y, z = tuple(self.xyz)
        return self.gd.integrate(n * Y_L(L, x, y, z, r2))

    def remove_moment(self, n, L, q=None):
        # Determine multipole moment
        if q is None:
            q = self.get_moment(n, L)

        # Don't do anything if moment is less than the tolerance
        if abs(q) < 1e-7:
            return 0.

        # Remove moment from input density
        n -= q * self.get_gauss(L)

        # Return correction
        return q * self.get_gauss_pot(L)


def gaussian_wave(r_vG, r0_v, sigma, k_v=None, A=None, dtype=float,
                  out_G=None):
    r"""Generates function values for atom-centered Gaussian waves.

    ::

                         _ _
        _            / -|r-r0|^2 \           _ _
      f(r) = A * exp( ----------- ) * exp( i k.r )
                     \ 2 sigma^2 /

    If the parameter A is not specified, the Gaussian wave is normalized::

                                                  oo
           /    ____        \ -3/2               /       _  2  2
      A = (    /    '        )        =>    4 pi | dr |f(r)|  r  = 1
           \ \/  pi   sigma /                    /
                                                   0

    Parameters:

    r_vG: ndarray
        Set of coordinates defining the grid positions.
    r0_v: ndarray
        Set of coordinates defining the center of the Gaussian envelope.
    sigma: float
        Specifies the spatial width of the Gaussian envelope.
    k_v: ndarray or None
        Set of reciprocal lattice coordinates defining the wave vector.
        An argument of None is interpreted as the gamma point i.e. k_v=0.
    A: float, complex or None
        Specifies the amplitude of the Gaussian wave. Normalizes if None.
    dtype: type, defaults to float
        Specifies the output data type. Only returns the real-part if float.
    out_G: ndarray or None
        Optional pre-allocated buffer to fill in values. Allocates if None.

    """
    if k_v is None:
        k_v = np.zeros(r0_v.shape)

    if A is None:
        # 4*pi*int(exp(-r^2/(2*sigma^2))^2 * r^2, r=0...infinity)
        # = sigma^3*pi^(3/2) = 1/A^2 -> A = (sqrt(Pi)*sigma)^(-3/2)
        A = 1 / (sigma * np.pi**0.5)**1.5

    if debug:
        assert is_contiguous(r_vG, float)
        assert is_contiguous(r0_v, float)
        assert is_contiguous(k_v, float)
        assert r_vG.ndim >= 2 and r_vG.shape[0] > 0
        assert r0_v.ndim == 1 and r0_v.shape[0] > 0
        assert k_v.ndim == 1 and k_v.shape[0] > 0
        assert (r_vG.shape[0], ) == r0_v.shape == k_v.shape
        assert sigma > 0

    if out_G is None:
        out_G = np.empty(r_vG.shape[1:], dtype=dtype)
    elif debug:
        assert is_contiguous(out_G)
        assert out_G.shape == r_vG.shape[1:]

    # slice_v2vG = [slice(None)] + [np.newaxis]*3
    # gw = lambda r_vG, r0_v, sigma, k_v, A=1/(sigma*np.pi**0.5)**1.5: \
    #    * np.exp(-np.sum((r_vG-r0_v[slice_v2vG])**2, axis=0)/(2*sigma**2)) \
    #    * np.exp(1j*np.sum(np.r_vG*k_v[slice_v2vG], axis=0)) * A
    cgpaw.utilities_gaussian_wave(A, r_vG, r0_v, sigma, k_v, out_G)
    return out_G



import pickle
import numpy as np
from math import pi
import ase.units
import os
import warnings

Hartree = ase.units.Hartree
Bohr = ase.units.Bohr


def load(fd):
    try:
        return pickle.load(fd, encoding='latin1')
    except TypeError:
        return pickle.load(fd)


class BuildingBlock:

    """ Module for using Linear response to calculate dielectric
    building block of 2D material with GPAW"""

    def __init__(self, filename, df, isotropic_q=None, nq_inf=10,
                 direction='x', qmax=None, txt='-', isotropic=True):
        """Creates a BuildingBlock object.

        filename: str
            used to save data file: filename-chi.npz
        df: DielectricFunction object
            Determines how linear response calculation is performed
        isotropic: bool
            If True, only q-points along one direction (1 0 0) in the
            2D BZ is included, thus assuming an isotropic material
        direction: 'x' or 'y'
            Direction used for isotropic q sampling.
        qmax: float
            Cutoff for q-grid. To be used if one wishes to sample outside the
            irreducible BZ. Only works for isotropic q-sampling.
        nq_inf: int
            number of extra q points in the limit q->0 along each direction,
            extrapolated from q=0, assuming that the head of chi0_wGG goes
            as q^2 and the wings as q.
            Note that this does not hold for (semi)metals!

        nq_inftot: int
            total number of extra q points for the q-> 0 limit.
            Equal to nq_inf for isotropic materials, and 2 * nq_inf otherwise

        """
        if isotropic_q is not None:
            warnings.warn('Keyword \'isotropic_q\' is deprecated and will be'
                          ' removed in the future. Use \'isotropic\' instead.',
                          DeprecationWarning)
            isotropic = isotropic_q

        assert isotropic, "Non-isotropic calculation" \
            + " temporarily turned-off until properly tested."
        if qmax is not None:
            assert isotropic
        self.filename = filename
        self.isotropic = isotropic
        self.nq_inf = nq_inf
        self.nq_inftot = nq_inf
        if not isotropic:
            self.nq_inftot *= 2
        if direction == 'x':
            qdir = 0
        elif direction == 'y':
            qdir = 1
        self.direction = direction

        self.df = df  # dielectric function object
        assert self.df.coulomb.truncation == '2D'
        self.wd = self.df.chi0calc.wd

        self.context = self.df.context.with_txt(txt)

        gs = self.df.gs
        kd = gs.kd
        self.kd = kd
        r = gs.gd.get_grid_point_coordinates()
        self.z_z = r[2, 0, 0, :]

        nw = len(self.wd)
        self.chiM_qw = np.zeros([0, nw])
        self.chiD_qw = np.zeros([0, nw])
        self.chiMD_qw = np.zeros([0, nw])
        self.chiDM_qw = np.zeros([0, nw])
        self.drhoM_qz = np.zeros([0, self.z_z.shape[0]])
        self.drhoD_qz = np.zeros([0, self.z_z.shape[0]])

        point_group_symmetries_scc = list(kd.symmetry.op_scc)
        z_inversion_matrix = np.diag([1, 1, -1])

        # If the material has z inversion symmetry, the off-diagonal elements
        # chiMD and chiDM are necessarily zero. In this case, it is not
        # necessary to calculate them.
        self.has_z_inversion_symmetry =\
            any(np.array_equal(z_inversion_matrix, sym)
                for sym in point_group_symmetries_scc)
        # First: choose all ibzq in 2D BZ
        from ase.dft.kpoints import monkhorst_pack
        from gpaw.kpt_descriptor import KPointDescriptor
        offset_c = 0.5 * ((kd.N_c + 1) % 2) / kd.N_c
        bzq_qc = monkhorst_pack(kd.N_c) + offset_c
        qd = KPointDescriptor(bzq_qc)
        qd.set_symmetry(gs.atoms, kd.symmetry)
        q_ibz_kc = qd.ibzk_kc
        rcell_cv = 2 * pi * np.linalg.inv(gs.gd.cell_cv).T
        if not isotropic:
            q_unsorted_kc = q_ibz_kc
        else:  # only use q along [1 0 0] or [0 1 0] direction.
            Nk = kd.N_c[qdir]
            qx = np.array(range(1, Nk // 2)) / float(Nk)
            q_unsorted_kc = np.zeros([Nk // 2 - 1, 3])
            q_unsorted_kc[:, qdir] = qx
            q = 0
            if qmax is not None:
                qmax *= Bohr
                qmax_v = np.zeros([3])
                qmax_v[qdir] = qmax
                q_c = q_unsorted_kc[-1]
                q_v = np.dot(q_c, rcell_cv)
                q = (q_v**2).sum()**0.5
                assert Nk % 2 == 0
                i = Nk / 2.0
                while q < qmax:
                    if i == Nk:  # omit BZ edge
                        i += 1
                        continue
                    q_c = np.zeros([3])
                    q_c[qdir] = i / Nk
                    q_unsorted_kc = np.append(q_unsorted_kc,
                                              q_c[np.newaxis, :], axis=0)
                    q_v = np.dot(q_c, rcell_cv)
                    q = (q_v**2).sum()**0.5
                    i += 1
        q_abs_unsorted_k = np.linalg.norm(q_unsorted_kc @ rcell_cv, axis=1)
        if isotropic:
            # extrapolate to half of smallest finite q
            q_cut = np.min(q_abs_unsorted_k) / 2
        else:
            q_cut = np.sort(q_abs_unsorted_k)[1]  # smallest finite q
        self.nq_cut = self.nq_inftot + 1

        # subscript q: q-points including q_infs
        q_infs_qv = np.zeros([len(q_unsorted_kc) + self.nq_inftot, 3])
        # x-direction:
        q_infs_qv[: self.nq_inftot, qdir] = \
            np.linspace(1e-05, q_cut, self.nq_inftot + 1)[:-1]
        if not isotropic:  # y-direction
            q_infs_qv[self.nq_inf:self.nq_inftot, 1] = \
                np.linspace(0, q_cut, self.nq_inf + 1)[1:]
        sort = np.argsort(q_abs_unsorted_k)
        # add q_inf to list
        q_sorted_kc = q_unsorted_kc[sort]
        self.q_qc = np.insert(q_sorted_kc, 0,
                              np.zeros([self.nq_inftot, 3]), axis=0)
        self.q_qv = self.q_qc @ rcell_cv
        self.q_qv += q_infs_qv
        self.q_abs_q = (self.q_qv**2).sum(axis=1)**0.5
        self.q_infs_qv = q_infs_qv
        self.complete = False
        self.last_q_idx = 0
        if self.load_chi_file():
            if self.complete:
                self.context.print('Building block loaded from file')
        self.context.comm.barrier()

    def calculate_building_block(self):
        if self.complete:
            return
        Nq = self.q_qc.shape[0]
        for current_q_idx in range(self.last_q_idx, Nq):
            self.save_chi_file(q_idx=current_q_idx)
            self.last_q_idx = current_q_idx
            q_c = self.q_qc[current_q_idx]
            q_inf = self.q_infs_qv[current_q_idx]
            if np.allclose(q_inf, 0):
                q_inf = None

            qcstr = '(' + ', '.join(['%.3f' % x for x in q_c]) + ')'
            self.context.print(
                'Calculating contribution from q-point #%d/%d, q_c=%s' % (
                    current_q_idx + 1, Nq, qcstr), flush=False)
            if q_inf is not None:
                qstr = '(' + ', '.join(['%.3f' % x for x in q_inf]) + ')'
                self.context.print('    and q_inf=%s' % qstr, flush=False)
            qpd, chi_wGG, wblocks = self.df.get_rpa_density_response(
                q_c=q_c, qinf_v=q_inf, direction=self.direction)
            self.context.print('calculated chi!')

            chiM_w, chiD_w, chiDM_w, chiMD_w, drhoM_z, drhoD_z = \
                self.get_chi_2D(qpd, chi_wGG)
            chiM_w = wblocks.all_gather(chiM_w)
            chiD_w = wblocks.all_gather(chiD_w)
            chiDM_w = wblocks.all_gather(chiDM_w)
            chiMD_w = wblocks.all_gather(chiMD_w)

            if self.context.comm.rank == 0:
                self.update_building_block(chiM_w[np.newaxis, :],
                                           chiD_w[np.newaxis, :],
                                           chiDM_w[np.newaxis, :],
                                           chiMD_w[np.newaxis, :],
                                           drhoM_z[np.newaxis, :],
                                           drhoD_z[np.newaxis, :])

        # Induced densities are not properly described in q-> 0 limit-
        # replace with finite q result:
        if self.context.comm.rank == 0:
            for n in range(Nq):
                if np.allclose(self.q_qc[n], 0):
                    self.drhoM_qz[n] = self.drhoM_qz[self.nq_cut]
                    self.drhoD_qz[n] = self.drhoD_qz[self.nq_cut]

        self.complete = True
        self.save_chi_file()

        return

    def update_building_block(self, chiM_qw, chiD_qw, chiDM_qw, chiMD_qw,
                              drhoM_qz, drhoD_qz):

        self.chiM_qw = np.append(self.chiM_qw, chiM_qw, axis=0)
        self.chiD_qw = np.append(self.chiD_qw, chiD_qw, axis=0)
        self.chiDM_qw = np.append(self.chiDM_qw, chiDM_qw, axis=0)
        self.chiMD_qw = np.append(self.chiMD_qw, chiMD_qw, axis=0)
        self.drhoM_qz = np.append(self.drhoM_qz, drhoM_qz, axis=0)
        self.drhoD_qz = np.append(self.drhoD_qz, drhoD_qz, axis=0)

    def get_chi_2D(self, qpd, chi_wGG):
        r"""Calculate the monopole and dipole contribution to the
        2D susceptibility chi_2D for single q-point q, defined as

        ::

          \chi^M_2D(q, \omega) = \int\int dr dr' \chi(q, \omega, r,r') \\
                              = L \chi_{G=G'=0}(q, \omega)
          \chi^D_2D(q, \omega) = \int\int dr dr' z \chi(q, \omega, r,r') z'
                               = 1/L sum_{G_z!=0, G_z'!=0} z_factor(G_z)
                               chi_{G_z,G_z'} z_factor(G_z'),
          \chi^DM_2D(q, \omega) = \int\int dr dr' z \chi(q, \omega,r,r')
                                = sum_{G_z != 0} z_factor(G_z) chi_{G_z,G'=0}
          \chi^MD_2D(q, \omega) = \int\int dr dr' \chi(q, \omega,r,r') z'
                            = sum_{G_z' != 0} chi_{0, G_z'} z_factor(G_z')^*
          Where z_factor(G_z) =  - i e^{i*G_z*z0}
          (L G_z cos(G_z L/2)-2 sin(G_z L/2))/G_z^2

        qpd: Single q-point descriptor
        chi_wGG: Susceptibility in PW basis
          """

        nw = chi_wGG.shape[0]
        z = self.z_z
        L = qpd.gd.cell_cv[2, 2]  # Length of cell in Bohr

        # XXX This seems like a bit dangerous assumption
        z0 = L / 2.  # position of layer
        chiM_w = np.zeros([nw], dtype=complex)
        chiD_w = np.zeros([nw], dtype=complex)
        chiDM_w = np.zeros([nw], dtype=complex)
        chiMD_w = np.zeros([nw], dtype=complex)
        drhoM_z = np.zeros([len(z)], dtype=complex)  # induced density
        drhoD_z = np.zeros([len(z)], dtype=complex)  # induced dipole density

        npw = chi_wGG.shape[1]
        G_Gv = qpd.get_reciprocal_vectors(add_q=False)

        Glist = []
        for iG in range(npw):  # List of G with Gx,Gy = 0
            if G_Gv[iG, 0] == 0 and G_Gv[iG, 1] == 0:
                Glist.append(iG)

        # If node lacks frequency points due to block parallelization then
        # return empty arrays
        if nw == 0:
            return chiM_w, chiD_w, chiDM_w, chiMD_w, drhoM_z, drhoD_z
        chiM_w = L * chi_wGG[:, 0, 0]
        drhoM_z += chi_wGG[0, 0, 0]
        for iG in Glist[1:]:
            G_z = G_Gv[iG, 2]
            qGr_R = np.inner(G_z, z.T).T
            factor = z_factor(z0, L, G_z)
            if not self.has_z_inversion_symmetry:
                # off-diagonal elements are non-zero only if
                # the material does not have z --> -z symmetry
                chiDM_w += factor * chi_wGG[:, iG, 0]
                chiMD_w += chi_wGG[:, 0, iG] * np.conjugate(factor)
            # Fourier transform to get induced density at \omega=0
            drhoM_z += np.exp(1j * qGr_R) * chi_wGG[0, iG, 0]
            for iG1 in Glist[1:]:
                G_z1 = G_Gv[iG1, 2]
                # integrate with z along both coordinates
                factor1 = z_factor(z0, L, G_z1, sign=-1)
                chiD_w[:] += 1. / L * factor * chi_wGG[:, iG, iG1] * \
                    factor1
                # induced dipole density due to V_ext = z
                drhoD_z[:] += 1. / L * np.exp(1j * qGr_R) * \
                    chi_wGG[0, iG, iG1] * factor1
        # Normalize induced densities with chi
        if nw != 0:
            drhoM_z /= chiM_w[0]
            drhoD_z /= chiD_w[0]

        """ Returns chi2D monopole and dipole, induced
        densities and z array (all in Bohr)
        """
        return chiM_w, chiD_w, chiDM_w, chiMD_w, drhoM_z, drhoD_z

    def save_chi_file(self, filename=None, q_idx=None):
        if q_idx is None:
            q_idx = self.last_q_idx
        if filename is None:
            filename = self.filename
        data = {'last_q': q_idx,
                'complete': self.complete,
                'isotropic_q': self.isotropic,  # old name for backwards compat
                'q_cs': self.q_qc,  # old name q_cs for backwards compatibility
                'q_vs': self.q_qv,  # old name q_vs for backwards compatibility
                'q_abs': self.q_abs_q,
                'omega_w': self.wd.omega_w,
                'chiM_qw': self.chiM_qw,
                'chiD_qw': self.chiD_qw,
                'chiDM_qw': self.chiDM_qw,
                'chiMD_qw': self.chiMD_qw,
                'z': self.z_z,
                'drhoM_qz': self.drhoM_qz,
                'drhoD_qz': self.drhoD_qz}

        if self.context.comm.rank == 0:
            np.savez_compressed(filename + '-chi.npz',
                                **data)
        self.context.comm.barrier()

    def load_chi_file(self):
        try:
            data = np.load(self.filename + '-chi.npz')
        except OSError:
            return False
        if (np.all(data['omega_w'] == self.wd.omega_w) and
            np.all(data['q_cs'] == self.q_qc) and
            np.all(data['z'] == self.z_z)):
            self.last_q_idx = data['last_q']
            self.complete = data['complete']
            self.chiM_qw = data['chiM_qw']
            self.chiD_qw = data['chiD_qw']
            self.drhoM_qz = data['drhoM_qz']
            self.drhoD_qz = data['drhoD_qz']
            if 'chiDM_qw' in data:
                self.chiDM_qw = data['chiDM_qw']
            else:
                self.chiDM_qw = np.zeros(self.chiM_qw.shape)
            if 'chiMD_qw' in data:
                self.chiMD_qw = data['chiMD_qw']
            else:
                self.chiMD_qw = np.zeros(self.chiM_qw.shape)

            return True
        else:
            return False

    def interpolate_to_grid(self, q_grid_q, w_grid_w,
                            q_grid=None, w_grid=None):
        """
        Parameters
        q_grid_q: in Ang. should start at q=0
        w_grid_w: in eV
        """

        if q_grid is not None or w_grid is not None:
            warnings.warn('\'q_grid\' and \'w_grid\' are deprecated and will'
                          ' be removed in a future version. Please use'
                          ' \'q_grid_q\' and \'w_grid_w\' instead',
                          DeprecationWarning)
            q_grid_q = q_grid
            w_grid_w = w_grid

        from scipy.interpolate import RectBivariateSpline
        from scipy.interpolate import interp1d
        from gpaw.response.frequencies import FrequencyGridDescriptor

        if not self.complete:
            self.calculate_building_block()

        q_grid_q = q_grid_q.copy() * Bohr
        w_grid_w = w_grid_w.copy() / Hartree

        # upper case subscripts refer to old grid.
        q_abs_Q = self.q_abs_q
        omega_W = self.wd.omega_w
        assert np.max(q_grid_q) <= np.max(q_abs_Q), \
            'q can not be larger that %1.2f Ang' % np.max(q_abs_Q / Bohr)
        assert np.max(w_grid_w) <= np.max(omega_W), \
            'w can not be larger that %1.2f eV' % \
            np.max(omega_W * Hartree)

        def spline(array, x_in, y_in, x_out, y_out):
            # interpolates a function from the regular grid (x_in, y_in)
            # to (x_out, y_out)
            # The shape of 'array' must be (len(x_in), len(y_in)).
            interpolator = RectBivariateSpline(x_in, y_in, array, s=0)
            return interpolator(x_out, y_out)

        def complex_spline(array, x_in, y_in, x_out, y_out):
            return spline(array.real, x_in, y_in, x_out, y_out)\
                + 1j * spline(array.imag, x_in, y_in, x_out, y_out)

        # chi monopole
        chiM_QW = self.chiM_qw

        omit_q0 = False
        if np.isclose(q_abs_Q[0], 0) and not np.isclose(chiM_QW[0, 0], 0):
            omit_q0 = True  # omit q=0 from interpolation
            q0_abs = q_abs_Q[0].copy()
            q_abs_Q[0] = 0.
            chi0_W = chiM_QW[0].copy()
            chiM_QW[0] = np.zeros_like(chi0_W)

        chiM_qw = complex_spline(chiM_QW, q_abs_Q, omega_W, q_grid_q, w_grid_w)
        if omit_q0:
            q_abs_Q[0] = q0_abs
            if np.isclose(q_grid_q[0], 0):
                yr = interp1d(omega_W, chi0_W.real)
                yi = interp1d(omega_W, chi0_W.imag)
                chi0_w = yr(w_grid_w) + 1j * yi(w_grid_w)
                chiM_qw[0] = chi0_w

        # chi dipole
        chiD_QW = self.chiD_qw
        chiD_qw = complex_spline(chiD_QW, q_abs_Q, omega_W,
                                 q_grid_q, w_grid_w)

        # chi off-diagonal
        if not self.has_z_inversion_symmetry:
            chiDM_QW = self.chiDM_qw
            chiDM_qw = complex_spline(chiDM_QW, q_abs_Q, omega_W,
                                      q_grid_q, w_grid_w)
            chiMD_QW = self.chiMD_qw
            chiMD_qw = complex_spline(chiMD_QW, q_abs_Q, omega_W,
                                      q_grid_q, w_grid_w)
        else:
            chiDM_qw = np.zeros((len(q_grid_q), len(w_grid_w)))
            chiMD_qw = np.zeros((len(q_grid_q), len(w_grid_w)))

        # drho monopole

        drhoM_Qz = self.drhoM_qz
        drhoM_qz = complex_spline(drhoM_Qz, q_abs_Q, self.z_z,
                                  q_grid_q, self.z_z)

        # drho dipole
        drhoD_Qz = self.drhoD_qz
        drhoD_qz = complex_spline(drhoD_Qz, q_abs_Q, self.z_z,
                                  q_grid_q, self.z_z)

        self.q_abs_q = q_grid_q
        self.wd = FrequencyGridDescriptor(w_grid_w)
        self.chiM_qw = chiM_qw
        self.chiD_qw = chiD_qw
        self.chiDM_qw = chiDM_qw
        self.chiMD_qw = chiMD_qw
        self.drhoM_qz = drhoM_qz
        self.drhoD_qz = drhoD_qz

        self.save_chi_file(filename=self.filename + '_int')

    def clear_temp_files(self):
        if not self.savechi0:
            comm = self.context.comm
            if comm.rank == 0:
                while len(self.temp_files) > 0:
                    filename = self.temp_files.pop()
                    os.remove(filename)


"""TOOLS"""


def check_building_blocks(BBfiles=None):
    """ Check that building blocks are on same frequency-
    and q- grid.

    BBfiles: list of str
        list of names of BB files
    """
    name = BBfiles[0] + '-chi.npz'
    data = np.load(name)
    try:
        q = data['q_abs'].copy()
        w = data['omega_w'].copy()
    except TypeError:
        # Skip test for old format:
        return True
    for name in BBfiles[1:]:
        data = np.load(name + '-chi.npz')
        if len(w) != len(data['omega_w']):
            return False
        elif not ((data['q_abs'] == q).all() and
                  (data['omega_w'] == w).all()):
            return False
    return True


def z_factor(z0, d, G, sign=1):
    factor = -1j * sign * np.exp(1j * sign * G * z0) * \
        (d * G * np.cos(G * d / 2.) - 2. * np.sin(G * d / 2.)) / G**2
    return factor


def z_factor2(z0, d, G, sign=1):
    factor = sign * np.exp(1j * sign * G * z0) * np.sin(G * d / 2.)
    return factor


def expand_layers(structure):
    newlist = []
    for name in structure:
        num = ''
        while name[0].isdigit():
            num += name[0]
            name = name[1:]
        try:
            num = int(num)
        except ValueError:
            num = 1
        for n in range(num):
            newlist.append(name)
    return newlist


def read_chi_wGG(name):
    """
    Read density response matrix calculated with the DielectricFunction
    module in GPAW.
    Returns frequency grid, gpaw.wavefunctions object, chi_wGG
    """
    fd = open(name, 'rb')
    omega_w, qpd, chi_wGG, q0, chi0_wvv = load(fd)
    nw = len(omega_w)
    nG = qpd.ngmax
    chi_wGG = np.empty((nw, nG, nG), complex)
    for chi_GG in chi_wGG:
        chi_GG[:] = load(fd)
    return omega_w, qpd, chi_wGG, q0



import pytest
import numpy as np
import pickle

from gpaw.lcao.tools import makeU, makeV
from gpaw import restart
from gpaw.lcao.pwf2 import LCAOwrap
from gpaw.mpi import world, rank, serial_comm

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.later
def test_lcao_pair_and_coulomb(gpw_files, in_tmp_dir):
    if rank == 0:
        atoms, calc = restart(gpw_files['h2_lcao_pair'],
                              txt=None, communicator=serial_comm)
        lcao = LCAOwrap(calc)
        H = lcao.get_hamiltonian()
        S = lcao.get_overlap()
        with open('lcao_pair_hs.pckl', 'wb') as fd:
            pickle.dump((H, S), fd, 2)
        # indices = get_bfi2(symbols, basis, scat)
        indices = range(2)
        lcao.get_xc(indices=indices).dump('lcao_pair_xc.pckl')
        lcao.get_Fcore(indices=indices).dump('lcao_pair_Fcore.pckl')
        w_wG = lcao.get_orbitals(indices=indices)
        P_awi = lcao.get_projections(indices=indices)
        with open('lcao_pair_w_wG__P_awi.pckl', 'wb') as fd:
            pickle.dump((w_wG, P_awi), fd, 2)

    world.barrier()
    makeU(gpw_files['h2_lcao_pair'],
          'lcao_pair_w_wG__P_awi.pckl',
          'lcao_pair_eps_q__U_pq.pckl',
          1.0e-5,
          dppname='lcao_pair_D_pp.pckl')

    world.barrier()
    makeV(gpw_files['h2_lcao_pair'],
          'lcao_pair_w_wG__P_awi.pckl',
          'lcao_pair_eps_q__U_pq.pckl',
          'lcao_pair_V_qq.pckl',
          'lcao_pair_V_qq.log',
          False)

    world.barrier()
    V_qq = np.load('lcao_pair_V_qq.pckl', allow_pickle=True)
    eps_q, U_pq = np.load('lcao_pair_eps_q__U_pq.pckl', allow_pickle=True)
    assert U_pq.flags.contiguous
    Usq_pq = U_pq * np.sqrt(eps_q)
    V_pp = np.dot(np.dot(Usq_pq, V_qq), Usq_pq.T.conj())
    V_pp_ref = np.array(
        [[15.34450177, 11.12669608, 11.12669608, 12.82934563],
         [11.12669608, 8.82280293, 8.82280293, 11.12669608],
         [11.12669608, 8.82280293, 8.82280293, 11.12669608],
         [12.82934563, 11.12669608, 11.12669608, 15.34450178]])
    assert abs(V_pp_ref - V_pp).max() == pytest.approx(0.0, abs=1.0e-5)



import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW, Mixer
from gpaw.mpi import world

# Place one atom next to cell boundary, then check that reuse_wfs
# works correctly when atom is subsequently displaced across the
# boundary, i.e., that the kpoint phases of the PAW correction
# are handled correctly when unprojecting/reprojecting the wavefunctions.


@pytest.mark.later  # Not implemented yet
def test_reuse_wfs_celldisp(in_tmp_dir):
    def check(reuse):
        atoms = molecule('H2')
        atoms.pbc = 1
        atoms.center(vacuum=1.5)
        atoms.positions -= atoms.positions[1]
        dz = 1e-2
        atoms.positions[:, 2] += dz

        calc = GPAW(mode='pw',
                    txt=f'gpaw-{reuse}.txt',
                    nbands=1,
                    experimental=dict(
                        reuse_wfs_method='paw' if reuse else None),
                    kpts=[[-0.3, 0.4, 0.2]],
                    symmetry='off',
                    mixer=Mixer(0.7, 5, 50.0))
        atoms.calc = calc

        for ctx in calc.icalculate(atoms):
            if ctx.niter == 2:
                # logerr1 = np.log10(calc.wfs.eigensolver.error)
                logerr1 = np.log10(ctx.wfs.eigensolver.error)

        atoms.positions[:, 2] -= 2 * dz

        for ctx in calc.icalculate(atoms, system_changes=['positions']):
            if ctx.niter == 2:
                logerr2 = np.log10(ctx.wfs.eigensolver.error)
                break

        if world.rank == 0:
            print(f'reuse={bool(reuse)}')
            print('logerr1', logerr1)
            print('logerr2', logerr2)
            gain = logerr2 - logerr1
            print('gain', gain)
        return logerr2

    noreuse_logerr = check(0)
    reuse_logerr = check(1)
    # Ref values: logerr=-4.8 without reuse_wfs and -6.1 with reuse_wfs
    assert reuse_logerr < -6.0, reuse_logerr
    assert reuse_logerr < noreuse_logerr - 1.2, (reuse_logerr, noreuse_logerr)



from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path

from gpaw.mpi import world


class CachedFilesHandler(ABC):
    """Base class for objects which handle the writing
    and caching of session-scoped pytest fixtures"""

    def __init__(self, folder: Path, outformat: str):
        self.folder = folder
        self.outformat = outformat

        self.cached_files = {}
        for file in folder.glob('*' + outformat):
            self.cached_files[file.name[:-len(outformat)]] = file

    def __getitem__(self, name: str) -> Path:
        if name in self.cached_files:
            return self.cached_files[name]

        filepath = self.folder / (name + self.outformat)

        lockfile = self.folder / f'{name}.lock'

        for _attempt in range(60):  # ~60s timeout
            file_exist = 0
            if world.rank == 0:
                file_exist = int(filepath.exists())
            file_exist = world.sum_scalar(file_exist)

            if file_exist:
                self.cached_files[name] = filepath
                return self.cached_files[name]

            try:
                with world_temporary_lock(lockfile):
                    work_path = filepath.with_suffix('.tmp' + self.outformat)
                    self._calculate_and_write(name, work_path)

                    # By now files should exist *and* be fully written, by us.
                    # Rename them to the final intended paths:
                    if world.rank == 0:
                        work_path.rename(filepath)

            except Locked:
                import time
                time.sleep(1)

        raise RuntimeError(f'{self.__class__.__name__} fixture generation '
                           f'takes too long: {name}.  Consider using pytest '
                           '--cache-clear if there are stale lockfiles, '
                           'else write faster tests.')

    @abstractmethod
    def _calculate_and_write(self, name, work_path):
        pass


class Locked(FileExistsError):
    pass


@contextmanager
def world_temporary_lock(path):
    if world.rank == 0:
        try:
            with temporary_lock(path):
                world.sum_scalar(1)
                yield
        except Locked:
            world.sum_scalar(0)
            raise
    else:
        status = world.sum_scalar(0)
        if status:
            yield
        else:
            raise Locked


@contextmanager
def temporary_lock(path):
    fd = None
    try:
        with path.open('x') as fd:
            yield
    except FileExistsError:
        raise Locked()
    finally:
        if fd is not None:
            path.unlink()



from math import exp, pi, sqrt
import numpy as np

from gpaw.gauss import Gauss
import pytest
from gpaw.utilities.folder import Folder, Lorentz, Voigt  # noqa

# Gauss and Lorentz functions


def test_spectrum():
    width = 0.5
    x = 1.5

    assert Gauss(width).get(x) == pytest.approx(
        exp(- x**2 / 2 / width**2) / sqrt(2 * pi) / width, abs=1.e-15)
    assert Gauss(width).fwhm == pytest.approx(
        width * np.sqrt(8 * np.log(2)), abs=1.e-15)
    assert Lorentz(width).get(x) == pytest.approx(
        width / (x**2 + width**2) / pi, abs=1.e-15)
    assert Lorentz(width).fwhm == pytest.approx(width * 2, abs=1.e-15)

    # folder function

    for func in [Gauss, Lorentz, Voigt]:
        folder = Folder(width, func(width).__class__.__name__)

        x = [0, 2]
        y = [[2, 0, 1], [1, 1, 1]]

        xl, yl = folder.fold(x, y, dx=.7)

        # check first value
        yy = np.dot(np.array(y)[:, 0], func(width).get(xl[0] - np.array(x)))
        assert yl[0, 0] == pytest.approx(yy, abs=1.e-15)
